/*
 * Copyright 2020 - 2021 IoT.bzh
 *
 * author: Valentin Lefebvre <valentin.lefebvre@iot.bzh>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            DEFINE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

#define _GNU_SOURCE

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            INCLUDE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

#include <ctype.h>
#include "ble-sensor-binding.h"

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            PRIVATE FUNCTIONS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

/**
 * discovered_device_handler - function call when a device is find by the gattlib library
 *
 * @param void *adapter: gattlib bluetooth adapter
 * @param const char *addr: string MAC address of sensor
 * @param const char *name: name of the device found (if any)
 * @param void *user_data: custom data sended (if any), these data are
 *  parametrized when you registered this callback
 *  function
 */
static void discovered_device_handler(void *adapter, const char *addr, const char *name, void *user_data) {
    BLE_gattlib_ble_sensor_t * sensor = (BLE_gattlib_ble_sensor_t *) user_data;

    if (strcmp(addr, sensor->macAddr) == 0) {
        sensor->state = SENSOR_FOUND;
        if (name) {
            sensor->name = (char *) name;
        }
        gattlib_adapter_scan_disable(adapter);
    }
}

/** 
 * find_sensor_ble - Scanning surrounding area for bluetooth low energy devices
 *
 * @param sensor sensor's information
 * @return ble_group_status_t
 */
static ble_group_status_t find_sensor_ble (BLE_gattlib_ble_sensor_t *sensor) {
    int ret = 0;
    void * adapter = NULL;

    // Open the Hardware adaptater
    ret = gattlib_adapter_open(NULL, &adapter);
    if (ret != GATTLIB_SUCCESS) {
        AFB_ERROR("Error when opening adaptater : error=%i!", ret);
        return BLE_ADAPTER_ERROR;
    }

    // Scan surrounding area
    sensor->state = SENSOR_NOT_FOUND;
    AFB_INFO("Starting to scan ...");
    ret = gattlib_adapter_scan_enable(adapter, discovered_device_handler, BLE_SCAN_TIMEOUT, (void *)sensor);
    if (ret != GATTLIB_SUCCESS) {
        AFB_ERROR("Error when scanning : error=%i!", ret);
        gattlib_adapter_close(adapter);
        return BLE_SCAN_ERROR;
    }

    // Check if Sensor has been found
    if (sensor->state == SENSOR_FOUND) {
        AFB_INFO("Sensor at %s has been found", sensor->macAddr);
        ret = BLE_SUCCESS ;
    } else {
        AFB_WARNING("Failed to find sensor");
        ret = BLE_DEVICE_ERROR;
    }
    gattlib_adapter_scan_disable(adapter);
    //gattlib_adapter_close(adapter);
    return ret;
}

/**
 * isValidMacAddress - Check if the MAC address passed as program parameter is valid
 *
 * @param const char *mac: MAC address passed by parameter
 * @return true / false
 */
static bool isValidMacAddress(const char *mac)
{
    int idx = 0;
    int nb_hexdigit = 0;
    int nb_separator = 0;

    while (mac[idx] != '\0') {
       if (isxdigit(mac[idx])) {
          ++nb_hexdigit;
       } else if (mac[idx] == ':' || mac[idx] == '-') {
          if (nb_hexdigit == 0 || nb_hexdigit / 2 - 1 != nb_separator) {
            break;
          }
          ++nb_separator;
       } else {
           nb_separator = -1;
       }
       ++idx;
    }
    if (nb_hexdigit == 12 && (nb_separator == 5 || nb_separator == 0)) {
        return true;
    }
    return false;
}


/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            PUBLIC FUNCTIONS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

void ble_gattlib_disconnect_handler(void * user_data) {
    assert(user_data);
    ble_sensor_t * sensor = (ble_sensor_t *) user_data;
    AFB_API_WARNING(sensor->api, "[disconnect_handler] The sensor %s/%s has been disconnected !", sensor->group->uid, sensor->uid);
    sensor->state = BLE_SENSOR_DISCONNECTED_STATE;
}

void ble_gattlib_connect_async_cb(gatt_connection_t* connection, void* user_data) {
    assert(user_data);
    ble_sensor_t * sensor = (ble_sensor_t *) user_data;

    sensor->connectingCount = 0;
    if (sensor->state == BLE_SENSOR_CONNECTED_STATE) return;

    if (!connection) {
        AFB_API_ERROR(sensor->api, "[connect_async_cb] Connection failed");
        sensor->state = BLE_SENSOR_DISCONNECTED_STATE;
        return;
    }
    AFB_API_INFO(sensor->api, "[connect_async_cb] Sensor %s/%s at %s is connected",
                        sensor->group->uid, sensor->uid, sensor->macAddr);
    // Getting version if function if defined
    if (sensor->group->functions->version_ptr_func) {
        AFB_API_INFO(sensor->api, "[ble_gattlib_connect_async_cb: %s/%s] Getting version ...", sensor->group->uid, sensor->uid);
        int ret = sensor->group->functions->version_ptr_func(sensor);
        if (ret < 0) {
            AFB_API_ERROR(sensor->api, "[connect_async_cb %s/%s] Error on getting version", sensor->group->uid, sensor->uid);
        }

    }
    sensor->state = BLE_SENSOR_CONNECTED_STATE;
    gattlib_register_on_disconnect(connection, ble_gattlib_disconnect_handler, sensor);
}

ble_group_status_t ble_gattlib_connect (ble_sensor_t *sensor, json_object * argsJ, BLEMode_t asyncMode) {
    BLE_gattlib_ble_sensor_t gatt_sensor;
    ble_group_status_t searching_status;
    int try =0;
    ble_group_status_t ret =0;

    // Change the MacAddress of the sensor of given
    if(argsJ && json_object_is_type(argsJ, json_type_string)) {
        if (!isValidMacAddress((char *)json_object_get_string(argsJ))){
            AFB_API_ERROR(sensor->api, "The data has to be an address mac in string");
            ret= -BLE_CONNECTION_ERROR;
            goto OnErrorExit;
        }
        sensor->macAddr = (char *)json_object_get_string(argsJ);
        AFB_API_INFO(sensor->api, "the Mac Address of the sensor %s/%s has been changed", sensor->group->uid, sensor->uid);
    }
    gatt_sensor.macAddr = sensor->macAddr;
    gatt_sensor.name = (char *)calloc(256, sizeof(char));

    // Try to find the sensor
    searching_status = find_sensor_ble(&gatt_sensor);
    if (searching_status != BLE_SUCCESS) {
        AFB_API_WARNING(sensor->api, "Fail to find sensor %s/%s!", sensor->group->uid, sensor->uid);
        ret = searching_status;
        goto OnErrorExit;
    }

    // Open the connection
    unsigned long conn_options = 0;
    conn_options = GATTLIB_CONNECTION_OPTIONS_LEGACY_BDADDR_LE_PUBLIC;
    conn_options |= GATTLIB_CONNECTION_OPTIONS_LEGACY_BT_SEC_HIGH;

    if (asyncMode == BLE_SENSOR_ASYNC_MODE) {
        sensor->gatt_connection = gattlib_connect_async(NULL, sensor->macAddr, conn_options, ble_gattlib_connect_async_cb, sensor);
    } else {
        while (!sensor->gatt_connection && try < 4) {
            AFB_API_DEBUG(sensor->api, "Try to connect: %i", try);
            sensor->gatt_connection = gattlib_connect(NULL, sensor->macAddr, conn_options);
            if (!sensor->gatt_connection) {
                AFB_API_DEBUG(sensor->api, "Try to connect with LE RANDOM option");
                conn_options = GATTLIB_CONNECTION_OPTIONS_LEGACY_BDADDR_LE_RANDOM;
                conn_options |= GATTLIB_CONNECTION_OPTIONS_LEGACY_BT_SEC_HIGH;
                sensor->gatt_connection = gattlib_connect(NULL, sensor->macAddr, conn_options);
            }
            if (sensor->gatt_connection) sensor->state = BLE_SENSOR_CONNECTED_STATE;
            try ++;
        }
    }

    if (!sensor->gatt_connection) {
        ret = -BLE_CONNECTION_ERROR;
        goto OnErrorExit;
    }
    if (gatt_sensor.name) free(gatt_sensor.name);
    return BLE_SUCCESS;

OnErrorExit:
    sensor->state = BLE_SENSOR_DISCONNECTED_STATE;
    return ret;
}

ble_group_status_t ble_gattlib_disconnect (gatt_connection_t * connection) {
    if (!connection) return BLE_CONNECTION_ERROR;
    return gattlib_disconnect(connection);
}

ble_group_status_t ble_gattlib_discover(ble_sensor_t *sensor, json_object ** outputJ) {
    if (!sensor->gatt_connection) return BLE_CONNECTION_ERROR;

    gattlib_primary_service_t* services;
	gattlib_characteristic_t* characteristics;
	int services_count, characteristics_count;
    char uuid_str[256], handleStart_str[5], handleEnd_str[5], prop_str[5], handle_str[7];
    int ret = BLE_SUCCESS, ret_gattlib = GATTLIB_SUCCESS;
    json_object *elemJ = NULL;
    json_object *servicesJ = json_object_new_array();
    json_object *characteristicsJ = json_object_new_array();

    // Discover Services
    ret_gattlib = gattlib_discover_primary(sensor->gatt_connection, &services, &services_count);
	if (ret_gattlib != GATTLIB_SUCCESS) {
		AFB_API_ERROR(sensor->api, "[ble_gattlib_discover] Discover primay for sensor %s/%s, error=%i", sensor->group->uid, sensor->uid, ret_gattlib);
        ret = -ret_gattlib;
		goto OnExit;
	}
    for (int index = 0; index < services_count; index++) {
		ret_gattlib = gattlib_uuid_to_string(&services[index].uuid, uuid_str, sizeof(uuid_str));
        if (ret_gattlib != GATTLIB_SUCCESS) {
            AFB_API_ERROR(sensor->api, "[ble_gattlib_discover] gattlib_uuid_to_string for sensor %s/%s, error=%i", sensor->group->uid, sensor->uid, ret_gattlib);
            ret = -ret_gattlib;
            goto FreeServices;
        }
        ret = snprintf(handleStart_str, sizeof(handleStart_str), "%02x", services[index].attr_handle_start);
        if (ret < 0) {
            AFB_API_ERROR(sensor->api, "[ble_gattlib_discover] snprintf for handle start: %i", ret);
            goto FreeServices;
        }
        ret = snprintf(handleEnd_str, sizeof(handleEnd_str), "%02x", services[index].attr_handle_end);
        if (ret < 0) {
            AFB_API_ERROR(sensor->api, "[ble_gattlib_discover] snprintf for handle end: %i", ret);
            goto FreeServices;
        }

        ret = wrap_json_pack(&elemJ, "{ss* ss* ss*}",
            "uuid", uuid_str,
            "attr_handle_start", handleStart_str,
            "attr_handle_stop", handleEnd_str);
        if (ret < 0) {
            AFB_API_ERROR(sensor->api, "[ble_gattlib_discover] Error occurs when packing a service");
            goto FreeServices;
        }
        ret = json_object_array_add(servicesJ, elemJ);
        if (ret < 0) {
            AFB_API_ERROR(sensor->api, "[ble_gattlib_discover] Error occurs when adding a service");
	        goto FreeServices;
        }
	}

    // Discover Characteristis
    ret_gattlib = gattlib_discover_char(sensor->gatt_connection, &characteristics, &characteristics_count);
	if (ret_gattlib != GATTLIB_SUCCESS) {
		AFB_API_ERROR(sensor->api, "[ble_gattlib_discover] Discover char for sensor %s/%s, error=%i", sensor->group->uid, sensor->uid, ret_gattlib);
        ret = -ret_gattlib;
		goto FreeCharacteristics;
	}
	for (int index = 0; index < characteristics_count; index++) {
		ret_gattlib = gattlib_uuid_to_string(&characteristics[index].uuid, uuid_str, sizeof(uuid_str));
        if (ret_gattlib != GATTLIB_SUCCESS) {
            AFB_API_ERROR(sensor->api, "[ble_gattlib_discover] gattlib_uuid_to_string characteristics for sensor %s/%s, error=%i", sensor->group->uid, sensor->uid, ret_gattlib);
            ret = -ret_gattlib;
            goto FreeCharacteristics;
        }
        ret = snprintf(prop_str, sizeof(prop_str), "%02x", characteristics[index].properties);
        if (ret < 0) {
            AFB_API_ERROR(sensor->api, "[ble_gattlib_discover] snprintf for characteristics properties: %i", ret);
            goto FreeCharacteristics;
        }
        ret = snprintf(handle_str, sizeof(handle_str), "%04x", characteristics[index].value_handle);
        if (ret < 0) {
            AFB_API_ERROR(sensor->api, "[ble_gattlib_discover] snprintf for characteristics value_handle: %i", ret);
            goto FreeCharacteristics;
        }

        ret = wrap_json_pack(&elemJ, "{ss* ss* ss*}",
            "uuid", uuid_str,
            "properties", prop_str,
            "value_handle", handle_str);
        if (ret < 0) {
            AFB_API_ERROR(sensor->api, "[ble_gattlib_discover] Error occurs when packing a characteristic");
            goto FreeCharacteristics;
        }

        ret = json_object_array_add(characteristicsJ, elemJ);
        if (ret < 0) {
            AFB_API_ERROR(sensor->api, "[ble_gattlib_discover] Error occurs when adding a characteristics");
            goto FreeCharacteristics;
        }
	}

    ret = wrap_json_pack(outputJ, "{so so}",
            "services", servicesJ,
            "characteristics", characteristicsJ);
    if (ret < 0) {
        AFB_API_ERROR(sensor->api, "[ble_gattlib_discover] Error occurs when packing response in JSON");
        goto FreeCharacteristics;
    }
    
    //return BLE_SUCCESS;
FreeCharacteristics:
    free(characteristics);

FreeServices:
    free(services);

OnExit:
    return ret;
}