/*
 * Copyright 2020 - 2021 IoT.bzh
 *
 * author: Valentin Lefebvre <valentin.lefebvre@iot.bzh>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            DEFINE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

#define _GNU_SOURCE

#ifndef CONTROL_CONFIG_PATH
    #define CONTROL_CONFIG_PATH "@CONTROL_CONFIG_PATH@"
#endif

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            INCLUDE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

#include "ble-sensor-binding.h"
#include <ctl-config.h>
#include <filescan-utils.h>

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            PROTOTYPE FUNCTION
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

static int BLE_Config(afb_api_t api, CtlSectionT *section, json_object *blesJ);
static void PingTest (afb_req_t request);
static void InfoSensors (afb_req_t request);

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            VARIABLES GLOBAL
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

static pthread_mutex_t g_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t cnd_scanning = PTHREAD_COND_INITIALIZER;
pthread_mutex_t mutex_scanning = PTHREAD_MUTEX_INITIALIZER;
bool is_scanning_run = false;

/**
 * ctrlSections - Array of callbacks linked to config file sections
 */
static CtlSectionT ctrlSections[] = {
    { .key = "plugins",         .loadCB = PluginConfig, .handle= ble_function_ptr_register},
    { .key = "onload",          .loadCB = OnloadConfig },
    { .key = "ble",             .loadCB = BLE_Config },
    { .key = NULL } /* marker for end of the array */
};

/**
 * CtrlApiVerb - Array of callbacks linked to verbs
 */
static afb_verb_t CtrlApiVerbs[] = {
    /* VERB'S NAME         FUNCTION TO CALL         SHORT DESCRIPTION */
    { .verb = "ping",     .callback = PingTest,         .info = "ble-sensor API ping test"},
    { .verb = "info",     .callback = InfoSensors,      .info = "BLE List sensors"},
    { .verb = NULL} /* marker for end of the array */
};

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            PRIVATE FUNCTIONS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

/** 
 * connectingThread - Take care about the connection of a CISS Bosh sensor
 */
static void *connectingThread (void *ctx) {
    assert(ctx);
    ble_sensor_thread_t *threadCtx = (ble_sensor_thread_t *) ctx;
    ble_sensor_t * sensor = (ble_sensor_t *) threadCtx->ctx;
	// unsigned long conn_options = 0;
    // conn_options = GATTLIB_CONNECTION_OPTIONS_LEGACY_BDADDR_LE_RANDOM;
    // conn_options |= GATTLIB_CONNECTION_OPTIONS_LEGACY_BDADDR_LE_PUBLIC;
    // conn_options |= GATTLIB_CONNECTION_OPTIONS_LEGACY_BT_SEC_HIGH;
    
    while(1) {
        pthread_mutex_lock(&g_mutex);

        switch (sensor->state)
        {
        case BLE_SENSOR_DISCONNECTED_STATE:
            sensor->state = BLE_SENSOR_CONNECTING_STATE;
            AFB_API_INFO(sensor->api, "[connecting thread: %s/%s] Connecting ...", sensor->group->uid, sensor->uid);
            //sensor->gatt_connection = gattlib_connect_async(NULL, sensor->macAddr, conn_options, ble_gattlib_connect_async_cb, sensor);
            sensor->gatt_connection = gattlib_connect(NULL, sensor->macAddr, GATTLIB_CONNECTION_OPTIONS_LEGACY_DEFAULT);
            if (sensor->gatt_connection == NULL) {
                AFB_API_DEBUG(sensor->api, "[connecting thread: %s/%s] at %s cannot be connected", sensor->group->uid, sensor->uid, sensor->macAddr);
                sensor->state = BLE_SENSOR_DISCONNECTED_STATE;
            } else {
                AFB_API_INFO(sensor->api, "[connecting thread: %s/%s] is connected",
                        sensor->group->uid, sensor->uid);

                if (sensor->group->functions->version_ptr_func) {
                    AFB_API_INFO(sensor->api, "[ble_gattlib_connect_async_cb: %s/%s] Getting version ...", sensor->group->uid, sensor->uid);
                    int ret = sensor->group->functions->version_ptr_func(sensor);
                    if (ret < 0) {
                        AFB_API_ERROR(sensor->api, "[connect_async_cb %s/%s] Error on getting version", sensor->group->uid, sensor->uid);
                    }
                }
                sensor->state = BLE_SENSOR_CONNECTED_STATE;
                gattlib_register_on_disconnect(sensor->gatt_connection, ble_gattlib_disconnect_handler, sensor);
            }

            break;

        case BLE_SENSOR_CONNECTING_STATE:
            if( sensor->connectingCount >= 2) {
                sensor->connectingCount = 0;
                AFB_API_INFO(sensor->api, "[connecting thread: %s/%s] still connecting...",
                    sensor->group->uid, sensor->uid);
            } else {
                sensor->connectingCount ++;
            }
            break;

        default:
            break;
        }
        pthread_mutex_unlock(&g_mutex);
        sleep(10);
    }

    AFB_WARNING("End of scanning thread");
    free(threadCtx);
    pthread_exit(NULL);
}

/**
 * discovered_device_handler - function call when a device is find by the gattlib library
 *
 * @param void *adapter: gattlib bluetooth adapter
 * @param const char *addr: string MAC address of sensor
 * @param const char *name: name of the device found (if any)
 * @param void *user_data: custom data sended (if any), these data are
 *  parametrized when you registered this callback
 *  function
 */
static void discovered_device_handler(void *adapter, const char *addr, const char *name, void *user_data) {
    userdata_t * userData = (userdata_t *) user_data;
	unsigned long conn_options = GATTLIB_CONNECTION_OPTIONS_LEGACY_BDADDR_LE_RANDOM;
    conn_options |= GATTLIB_CONNECTION_OPTIONS_LEGACY_BT_SEC_LOW;
    int ret = 0;
    // Loop on every sensor
    for (int idxGrp = 0; idxGrp< userData->nbrGroups; idxGrp++) {
        ble_group_t * group = &userData->groups[idxGrp];
        for (int idxSensor = 0; idxSensor < group->nbrSensors; idxSensor++) {
            ble_sensor_t * sensor = &group->sensors[idxSensor];
            if (strcmp(addr, sensor->macAddr) == 0) {
                if(!sensor->found) {
                    sensor->found = true;
                    AFB_API_DEBUG(sensor->api, "[discovered_sensor] the sensor %s (%s) has been found", name, addr);

                    ble_sensor_thread_t * threadCtx = calloc(1, sizeof(ble_sensor_thread_t));
                    threadCtx->ctx = (void *)sensor;
                    threadCtx->api = sensor->api;
                    ret = pthread_create(&sensor->threadID, NULL, connectingThread, threadCtx);
                    if (ret < 0) {
                        AFB_API_ERROR(sensor->api, "[discovered_sensor] Error to create connecting thread");
                    }
                }
                return;
            }
        }
    }
}

/** 
 * scanningThread - Take care about the scan of BLuetooth area
 */
static void *scanningThread (void *ctx) {
    assert(ctx);
    ble_sensor_thread_t * threadCtx = (ble_sensor_thread_t *) ctx;
    userdata_t * userData = (userdata_t *) threadCtx->ctx;
    int ret = 0;

    // Open the Hardware adaptater and scan the bluetooth area
    void * adapter = NULL;
    ret = gattlib_adapter_open(NULL, &adapter);
    if (ret != GATTLIB_SUCCESS) {
        AFB_ERROR("Error when opening adaptater : error=%i !", ret);
        pthread_exit(NULL);
    }

    while(1) {
    // Scan with a Timeout
        pthread_mutex_lock(&mutex_scanning);
        pthread_cond_wait(& cnd_scanning, &mutex_scanning);
        pthread_mutex_unlock(&mutex_scanning);
        AFB_NOTICE("Scan starting in background");
        is_scanning_run = true;
        pthread_mutex_lock(&g_mutex);
        ret = gattlib_adapter_scan_enable(adapter, discovered_device_handler, BLE_SCAN_TIMEOUT, (void *)userData);
        if (ret != GATTLIB_SUCCESS) {
            ret = gattlib_adapter_scan_disable(adapter);
            if (ret != GATTLIB_SUCCESS) {
                AFB_ERROR("[scanning thread] Error to disable scan error: %i", ret);
            }
            ret = gattlib_adapter_scan_enable(adapter, discovered_device_handler, BLE_SCAN_TIMEOUT, (void *)userData);
            if (ret != GATTLIB_SUCCESS) {
                AFB_ERROR("[scanning thread] Error to enable scan error: %i", ret);
                goto OnExit;
            }
        }

        // Disable scan
        ret = gattlib_adapter_scan_disable(adapter);
        if (ret != GATTLIB_SUCCESS) {
            AFB_ERROR("[scanning thread] Error to disable scan error: %i", ret);
        }
        pthread_mutex_unlock(&g_mutex);
        is_scanning_run = false;
        AFB_NOTICE("Scan stoping in background");

        //sleep(20);
    }

OnExit:
    // Close the adapter
    ret = gattlib_adapter_scan_disable(adapter);
    if (ret != GATTLIB_SUCCESS) {
        AFB_ERROR("[scanning thread] Error to disable scan error: %i", ret);
    }
    ret = gattlib_adapter_close(adapter);
    if (ret != GATTLIB_SUCCESS) {
        AFB_ERROR("[scanning thread] Error to disable scan");
    }

    AFB_WARNING("End of scanning thread");
    pthread_exit(NULL);
}


/**
 * stateCB - Function call due to reader Timer Take care about the connection
 * 
 *  @param TimerHandleT *timer
 */
static int stateCB(TimerHandleT *timer) {
    assert(timer);
    userdata_t * userdata = (userdata_t *) timer->context;
    int ret = 0;
    bool needScanning = false;
	// unsigned long conn_options = 0;

    // Loop on every ble sensor
    for (int indexGrp = 0; indexGrp < userdata->nbrGroups; indexGrp ++) {
        ble_group_t * bleGroup = &userdata->groups[indexGrp];
        for (int indexSen =0; indexSen < bleGroup->nbrSensors; indexSen ++) {
            ble_sensor_t * bleSensor = &bleGroup->sensors[indexSen];

            // Check the connection state of the sensor
            switch (bleSensor->state)
            {

            case BLE_SENSOR_CONNECTED_STATE:
                if (bleSensor->group->functions->data_streaming_ptr_func) {
                    AFB_API_INFO(timer->api, "[state sensor: %s/%s] Starting the data streaming ...", bleGroup->uid, bleSensor->uid);
                    
                    ret = bleGroup->functions->data_streaming_ptr_func(bleSensor);
                    if (ret < 0) {
                        AFB_API_ERROR(timer->api, "[state sensor: %s/%s] Error on dataStreaming CB", bleGroup->uid, bleSensor->uid);
                    }
                } else {
                    AFB_API_DEBUG(timer->api, "[state sensor %s/%s] No dataStreaming Callbacks", bleGroup->uid, bleSensor->uid);
                } 
                break;

            case BLE_SENSOR_ERROR_STATE:
                ret = gattlib_disconnect(bleSensor->gatt_connection);
                if (ret != GATTLIB_SUCCESS) {
                    AFB_API_ERROR(bleSensor->api, "Error to disconnect sensor %s/%s -> Error: %i", bleSensor->group->uid, bleSensor->uid, ret);
                }
                pthread_cancel(bleSensor->threadID);
                pthread_join(bleSensor->threadID, NULL);
                bleSensor->found = false;
                bleSensor->state = BLE_SENSOR_DISCONNECTED_STATE;
                sleep(2);
                break;
    
            default:
                break;
            }
            needScanning = needScanning | !bleSensor->found;
        }
    }

    if (!is_scanning_run && needScanning) {
        pthread_mutex_lock(&mutex_scanning);
        pthread_cond_signal(&cnd_scanning);
        pthread_mutex_unlock(&mutex_scanning);
    }

    return 1;
}

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            VERBS FUNCTIONS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

/**
 * PingTest - Callback of the ping verb
 * 
 * @param request afb request
 */
static void PingTest (afb_req_t request) {
    static int count=0;
    char response[32];
    json_object *queryJ =  afb_req_json(request);

    snprintf (response, sizeof(response), "Pong=%d", count++);
    AFB_API_NOTICE(request->api, "ble-sensor:ping count=%d query=%s", count, json_object_get_string(queryJ));
    afb_req_success_f(request,json_object_new_string(response), NULL);

    return;
}

/**
 * InfoSensors - Callback of the info verb
 * 
 * @param request afb request
 */
static void InfoSensors (afb_req_t request) {
    json_object *elemJ;
    int ret = 0, index;
    int verbose = 0;

    afb_api_t api = afb_req_get_api(request);
    CtlConfigT * ctlConfig = (CtlConfigT*)afb_api_get_userdata(api);
    userdata_t * userdata = (userdata_t *) ctlConfig->external;
    ble_group_t * ble = userdata->groups;
    json_object * queryJ =  afb_req_json(request);
    json_object * responseJ = json_object_new_array();

    if (json_object_is_type(queryJ, json_type_object)) {
        ret = wrap_json_unpack(queryJ, "{s?i !}", "verbose", &verbose);
        if (ret < 0) {
            afb_req_fail_f(request, "FAILED", "[InfoSensors] Invalid 'json query' query=%s", json_object_get_string(queryJ));
            return;
        }
    }

    /* Loop on every BLE group */
    if (verbose) {
        for (index=0; index < userdata->nbrGroups; index++) {
            switch (verbose)
            {
            default:
            case 1:
                ret = wrap_json_pack(&elemJ, "{ss ss}",
                        "uid", ble[index].uid,
                        "info", ble[index].info);
                if (ret < 0 ) {
                    AFB_API_ERROR(api, "[InfoSensor] Unable to wrap");
                    goto OnErrorExit;
                }
                break;
            case 2:
                ret = wrap_json_pack(&elemJ, "{ss ss si}",
                        "uid", ble[index].uid,
                        "info", ble[index].info,
                        "nbrSensor", ble[index].nbrSensors);
                if (ret < 0 ) {
                    AFB_API_ERROR(api, "[InfoSensor] Unable to wrap");
                    goto OnErrorExit;
                }
                break;
            case 3:
                ret = wrap_json_pack(&elemJ, "{ss ss}",
                        "uid", ble[index].uid,
                        "info", ble[index].info,
                        "nbrSensor", ble[index].nbrSensors);
                if (ret < 0 ) {
                    AFB_API_ERROR(api, "[InfoSensor] Unable to wrap");
                    goto OnErrorExit;
                }
                break;
            }
            ret = json_object_array_add(responseJ, elemJ);
            if (ret < 0) {
                AFB_API_ERROR(api, "[InfoSensor] Unable to add elem ble %s", ble[index].uid);
                goto OnErrorExit;
            }
        }
    } else {
        // build global info page for developper dynamic HTML5 page
        json_object * globalJ, * bleJ, * blesJ, * sensorsJ, * actionsJ, * usageJ;
        ret = wrap_json_pack (&globalJ, "{ss ss* ss* ss*}",
                "uid", ctlConfig->api,
                "info", ctlConfig->info,
                "version", ctlConfig->version,
                "author", ctlConfig->author);
        if (ret < 0) {
            AFB_API_ERROR(api, "[InfoSensor] Unable to wrap globalJ");
            goto OnErrorExit;
        }

        blesJ = json_object_new_array();
        for (index = 0; index < userdata->nbrGroups; index++) {
            // Prepare array to hold all Sensor verbs
            sensorsJ = json_object_new_array();
            ret = wrap_json_pack(&actionsJ, "[s s s]", "info", "connect", "disconnect");
            if (ret < 0) {
                AFB_API_ERROR(api, "[InfoSensors] Fail to wrap Json actionsJ metadata for");
                goto OnErrorExit;
            }
            ret = wrap_json_pack (&usageJ, "{so, si}", "action", actionsJ, "verbose", 3);
            if (ret < 0) {
                AFB_API_ERROR(api, "[InfoSensors] Fail to wrap Json usageJ metadata for");
                goto OnErrorExit;
            }

            // Create group object with infos and sensors list
            ret = ble_glue_sensors_info(&ble[index], 3, sensorsJ);
            if (ret < 0) {
                AFB_API_ERROR (api, "[InfoSensors] Fail to create ble group=%s", ble[index].uid);
                goto OnErrorExit;
            }
            ret = wrap_json_pack(&bleJ, "{ss ss* so}",
                    "uid", ble[index].uid,
                    "info", ble[index].info,
                    "verbs", sensorsJ);
            if (ret < 0) {
                AFB_API_ERROR (api, "[InfoSensors] Fail to wrap json sensors info ble=%s", ble[index].uid);
                goto OnErrorExit;
            }
            ret = json_object_array_add(blesJ, bleJ);
            if (ret < 0) {
                AFB_API_ERROR(api, "[InfoSensor] Unable to add elem ble %s", ble[index].uid);
                goto OnErrorExit;
            }
        }

        ret = wrap_json_pack(&responseJ, "{so so}",
                "metadata", globalJ,
                "groups", blesJ);
        if (ret < 0) {
            AFB_API_ERROR(api, "[InfoSensors] Failed to wrap Json Global response");
            goto OnErrorExit;
        }
    }

    afb_req_success(request, responseJ, NULL);
    return;

OnErrorExit:
    afb_req_fail_f(request, "FAILED", "[InfoSensors] Failed to get Info");
    return;
}

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            CONTROLLER FUNCTIONS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

/**
 * CtrlLoadStaticVerbs - Load all Statics verbs
 * 
 * @param api afb api
 * @param verbs pointer to the list of verbs with their callbacks
 * @param vcbdata context data
 */
static int CtrlLoadStaticVerbs (afb_api_t api, afb_verb_t *verbs, void *vcbdata) {
    int errcount=0;

    for (int idx=0; verbs[idx].verb; idx++) {
        errcount+= afb_api_add_verb(api, CtrlApiVerbs[idx].verb, CtrlApiVerbs[idx].info, CtrlApiVerbs[idx].callback, vcbdata, 0, 0, 0);
    }

    return errcount;
};

/**
 * Sensor_DynRequest - CAll dynamically the function which take cares about request for Sensor
 * 
 * @param request afb request
 */
static void Sensor_DynRequest(afb_req_t request) {
    // retrieve action handle from request and execute the request
    json_object *queryJ = afb_req_json(request);
    ble_sensor_t* sensor = (ble_sensor_t *) afb_req_get_vcbdata(request);

    ble_glue_sensor_request(request, sensor, queryJ);
}

/**
 * BLE_DynRequest - CAll dynamically the function which take cares about request for BLE
 * 
 * @param request afb request
 */
static void BLE_DynRequest(afb_req_t request) {
    // retrieve action handle from request and execute the request
    json_object *queryJ = afb_req_json(request);
    ble_group_t* ble= (ble_group_t*) afb_req_get_vcbdata(request);
    ble_glue_group_request(request, ble, queryJ);
}


/**
 * Sensor_LoadOne - Check the config file to Load a sensor
 * 
 * @param api afb_api_t 
 * @param ble ptr to the ble_group_t
 * @param sensor ptr to the ble_sensor_t
 * @param sensorJ Json of the sensor config.
 * @returns 0 if success negative if not
 */
static int Sensor_LoadOne(afb_api_t api, ble_group_t * ble, ble_sensor_t * sensor, json_object * sensorJ, int count) {
    int err = 0;
    json_object *argJ=NULL;
    const char* mode = NULL;

    // Should already be allocated
    assert (sensorJ);

    memset(sensor, 0, sizeof(ble_sensor_t));
    sensor->group = ble;

    // Unwrap all necessary information from conf file
    err = wrap_json_unpack(sensorJ, "{ss,s?s,s?s,s?s,s?o,s?o}",
                "uid", &sensor->uid,
                "info", &sensor->info,
                "macAddr", &sensor->macAddr,
                "mode", &mode,
                "userData", &argJ,
                "sample", &sensor->sample);
    if (err) {
        AFB_API_ERROR(api, "Sensor_LoadOne: Fail to Parse Sensor configs");
        return -1;
    }

    // Keep Sample as object when defined
    // if (sensor->sample) json_object_get (sensor->sample);

    // Store the api into the sensor object
    sensor->api = api;
    
    // If defined call function init callback
    if (ble->functions->init_ptr_func) {
        err = ble->functions->init_ptr_func(sensor, argJ);
        if (err) {
            AFB_API_ERROR(api, "Sensor_LoadOne: fail during init callback for sensor=%s", sensor->uid);
            return -1;
        }
    }
    
    // Create Verb
    err = asprintf ((char **) &sensor->apiverb, "%s/%s", ble->prefix, sensor->uid);
    err = afb_api_add_verb(api, sensor->apiverb, sensor->info, Sensor_DynRequest, sensor, NULL, 0, 0);
    if (err) {
        AFB_API_ERROR(api, "Sensor_LoadOne: fail to register API verb = %s", sensor->apiverb);
        return -1;
    }

    // Init some usefull variables
    sensor->gatt_connection = NULL;
    sensor->state = BLE_SENSOR_DISCONNECTED_STATE;
    sensor->connectingCount = 0;
    sensor->noDataCount = 0;
    sensor->version = json_object_new_object();
    sensor->found = false;

    return 0;
}

/**
 * BLE_LoadOne - Check the config file to Load a type
 *  of sensor, create dynamically a verb
 * 
 * @param api afb_api_t 
 * @param ble ptr of ble_group_t
 * @param bleJ Json of the ble config.
 * @returns: errors
 */
static int BLE_LoadOne(afb_api_t api, ble_group_t * ble, json_object * bleJ) {
    int err = 0;
    json_object *sensorsJ;
    const char *format = NULL;

    // Should already be allocated
    assert(bleJ);
    assert(api);

    memset(ble, 0, sizeof(ble_group_t));
    err = wrap_json_unpack(bleJ, "{ss,s?s,s?s,s?s,s?o}",
            "uid", &ble->uid,
            "info", &ble->info,
            "prefix", &ble->prefix,
            "format", &format,
            "sensors", &sensorsJ);
    if (err) {
        AFB_API_ERROR(api, "Fail to parse ble sensor JSON : (%s)", json_object_to_json_string(bleJ));
        goto OnErrorExit;
    }

    // If not API prefix let's use BLE uid
    if (!ble->prefix) ble->prefix = ble->uid;

    err=asprintf ((char **) &ble->adminapi, "%s/%s", ble->prefix, "admin");
    // Create verb
    err = afb_api_add_verb(api, ble->adminapi, ble->info, BLE_DynRequest, ble, NULL, 0, 0);
    if (err) {
        AFB_API_ERROR(api, "BLE_LoadOne: fail to register API uid=%s info=%s", ble->uid, ble->info);
        goto OnErrorExit;
    }

    // Find function Callback
    ble->functions = ble_function_ptr_find(api, format);
    if (!ble->functions) goto FunctionsErrorExit;

    // Loop on sensors
    if (json_object_is_type(sensorsJ, json_type_array)) {
        ble->nbrSensors = (int) json_object_array_length(sensorsJ);
        ble->sensors = (ble_sensor_t *) calloc(ble->nbrSensors +1, sizeof(ble_sensor_t));

        for (int index = 0; index < ble->nbrSensors; index++) {
            json_object * sensorJ = json_object_array_get_idx(sensorsJ, index);
            err = Sensor_LoadOne(api, ble, &ble->sensors[index], sensorJ, index);
            if (err) goto OnErrorExit;
        }
    } else {
        ble->sensors = (ble_sensor_t *) calloc(2, sizeof(ble_sensor_t));
        err = Sensor_LoadOne(api, ble, &ble->sensors[0], sensorsJ, 0);
        if (err) goto OnErrorExit;
    }

    return 0;

FunctionsErrorExit:
    AFB_API_ERROR(api, "BLE_LoadOne: missing/invalid Format Type=%s JSON=%s", ble->uid, json_object_to_json_string(bleJ));
    goto OnErrorExit;

OnErrorExit:
    return -1;
}

/**
 * BLE_Config - Callback of the controller section "ble"
 *  Load and store into struct the API config
 * 
 * @param api afb_api_t 
 * @param section ptr of CtlSectionT
 * @param bleJ Json of the ble config
 * @returns 0 if success, negative if not
 */
static int BLE_Config(afb_api_t api, CtlSectionT *section, json_object *blesJ) {
    int ret = 0;
    CtlConfigT * ctlConfig = (CtlConfigT*)afb_api_get_userdata(api);
    userdata_t * userdata = (userdata_t *) ctlConfig->external;
    // Everything is done during initial config call
    if (!blesJ) return 0;

    if (json_object_is_type(blesJ, json_type_array)) {
        userdata->nbrGroups = (int) json_object_array_length(blesJ);
        userdata->groups = (ble_group_t *) calloc(userdata->nbrGroups, sizeof(ble_group_t));

        for (int index = 0; index < userdata->nbrGroups; index++) {
            json_object * bleJ = json_object_array_get_idx(blesJ, index);
            ret = BLE_LoadOne(api, &userdata->groups[index], bleJ);
            if (ret < 0) goto OnErrorExit;
        }
    } else {
        userdata->nbrGroups = 1;
        userdata->groups = (ble_group_t *) calloc(userdata->nbrGroups, sizeof(ble_group_t));
        ret = BLE_LoadOne(api, &userdata->groups[0], blesJ);
        if (ret < 0) goto OnErrorExit;
    }

    // add static controls verbs
    ret = CtrlLoadStaticVerbs (api, CtrlApiVerbs, (void*) userdata->groups);
    if (ret < 0) {
        AFB_API_ERROR(api, "BLE_Config fail to Load static API verbs");
        goto OnErrorExit;
    }


    //sensor_thread_t * threadCtx = calloc(1, sizeof(sensor_thread_t));
    //threadCtx->sensor = sensor;
    ble_sensor_thread_t * threadCtx = calloc(1, sizeof(ble_sensor_thread_t));
    threadCtx->ctx = (void *)userdata;
    threadCtx->api = api;
    threadCtx->threadID = 0;
    ret = pthread_create(&threadCtx->threadID, NULL, scanningThread, threadCtx);
    if (ret < 0) {
        AFB_API_ERROR(api, "[BLE_Config] Error to create scanning thread");
        return -1;
    }
    // ret = pthread_create(&threadCtx->threadID, NULL, connectingThread, threadCtx);
    // if (ret < 0) {
    //     AFB_API_ERROR(api, "[BLE_Config] Error to create connecting thread");
    //     return -1;
    // }

    // Start Timer witch check state of each Sensor
    if (!userdata->groups->timer) {
        userdata->groups->timer = (TimerHandleT *) calloc(1, sizeof(TimerHandleT));
        userdata->groups->timer->count = -1; // run forever
        userdata->groups->timer->delay = SCAN_DELAY_AUTO * 1000;
    }
    TimerEvtStart(api, userdata->groups->timer, stateCB, userdata);
    stateCB(userdata->groups->timer);

    return 0;

OnErrorExit:
    AFB_API_ERROR(api, "Fail to Initiate BLE check Json Config");
    return -1;
}

/**
 * CtrlInitOneApi - Init a single API
 * 
 * @param api afb_api_t 
 * 
 * @returns 0 if success, negative in error
 */
static int CtrlInitOneApi(afb_api_t api) {
    int err = 0;

    // retrieve section config from api handle
    CtlConfigT* ctlConfig = (CtlConfigT*)afb_api_get_userdata(api);

    err = CtlConfigExec(api, ctlConfig);
    if (err) {
        AFB_API_ERROR(api, "Error at CtlConfigExec step");
        return err;
    }

    return err;
}

/**
 * CtrlLoadOneApi - Load a single API
 * 
 * @param vcbdata ptr of CtlConfigT
 * @param api afb_api_t 
 * 
 * @return errors
 */
static int CtrlLoadOneApi(void* vcbdata, afb_api_t api) {
    CtlConfigT* ctlConfig = (CtlConfigT*)vcbdata;

    // Init external DAta for the API
    userdata_t *userData = calloc(1, sizeof(userdata_t));
    userData->nbrGroups = 0;
    ctlConfig->external = (void *) userData;

    // save closure as api's data context
    afb_api_set_userdata(api, ctlConfig);

    // load section for corresponding API
    int error = CtlLoadSections(api, ctlConfig, ctrlSections);

    // init and seal API function
    afb_api_on_init(api, CtrlInitOneApi);
    afb_api_seal(api);

    return error;    
}

/**
 * afbBindingEntry - Entry point of the binding
 * 
 * @param api afb_api_t 
 * @return status
 */
int afbBindingEntry(afb_api_t api) {
    int status = 0;
    char *searchPath = NULL;
    const char *envConfig = NULL;
    afb_api_t handle;

    AFB_API_NOTICE(api, "Controller in afbBindingEntry");

    envConfig = getenv("CONTROL_CONFIG_PATH");
    if (!envConfig){
        AFB_API_NOTICE(api, "Using default environnement config path : %s", CONTROL_CONFIG_PATH);
        envConfig = CONTROL_CONFIG_PATH;
    }
    else AFB_API_NOTICE(api, "Found environnement config path : %s", envConfig);


    status=asprintf (&searchPath,"%s:%s", envConfig, GetBindingDirPath(api));
    AFB_API_NOTICE(api, "Json config directory : %s", searchPath);

    const char* prefix = "control";
    const char* configPath = CtlConfigSearch(api, searchPath, prefix);
    if (!configPath) {
        AFB_API_ERROR(api, "afbBindingEntry: No %s-%s* config found in %s ", prefix, GetBinderName(), searchPath);
        status = ERROR;
        free(searchPath);
        return status;
    }

    AFB_API_NOTICE(api, "api will be using config : %s", configPath);

    // load config file and create API
    CtlConfigT* ctrlConfig = CtlLoadMetaData(api, configPath);
    if (!ctrlConfig) {
        AFB_API_ERROR(api, "afbBindingEntry No valid control config file in:\n-- %s", configPath);
        status = ERROR;
        free(searchPath);
        return status;
    }

    AFB_API_NOTICE(api, "Controller API='%s' info='%s'", ctrlConfig->api, ctrlConfig->info);

    // create one API per config file (Pre-V3 return code ToBeChanged)
    handle = afb_api_new_api(api, ctrlConfig->api, ctrlConfig->info, 1, CtrlLoadOneApi, ctrlConfig);
    status = (handle) ? 0 : -1;

    return status;
   
}