/*
 * Copyright 2020 - 2021 IoT.bzh
 *
 * author: Valentin Lefebvre <valentin.lefebvre@iot.bzh>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            DEFINE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

#ifndef _BLE_SENSOR_BINDING_INCLUDE_
#define _BLE_SENSOR_BINDING_INCLUDE_

#define SCAN_DELAY_AUTO   5
#define  AFB_BINDING_VERSION 3
#ifndef ERROR
  #define ERROR -1
#endif

/* Other usefull constants for the rest of the program :
 *
 * - UUID_LENGTH : size of a string UUID, it avoids multiple call of "strlen"
 *                 when we want to transform raw uuid to string uuid
 *
 * - BLE_SCAN_TIMEOUT : defines the time of each scans
 */
#define UUID_LENGTH 36
#define BLE_SCAN_TIMEOUT 4

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            INCLUDE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

// usefull classical include
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <stdbool.h>
#include <unistd.h>
#include <pthread.h>

#include <afb/afb-binding.h>
#include <afb-timer.h>
#include <wrap-json.h>
#include <ctl-config.h>

#include <gattlib.h>


/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            ENUM
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

typedef enum BLEState_e{
  BLE_SENSOR_CONNECTED_STATE=0,
  BLE_SENSOR_CONNECTING_STATE,
  BLE_SENSOR_DISCONNECTED_STATE,
  BLE_SENSOR_CONFIG_STATE,
  BLE_SENSOR_STREAMING_STATE,
  BLE_SENSOR_ERROR_STATE
}BLEState_t;

typedef enum BLEMode_e{
  BLE_SENSOR_MANUAL_MODE=0,
  BLE_SENSOR_AUTO_MODE,
  BLE_SENSOR_ASYNC_MODE,
  BLE_SENSOR_NON_ASYNC_MODE
}BLEMode_t;

typedef enum ble_group_status_e {
    BLE_SUCCESS = 0,
    BLE_NOT_CONNECTED,
    BLE_WRONG_STATE,
    BLE_CONFIG_ERROR,
    BLE_ADAPTER_ERROR,
    BLE_SCAN_ERROR,
    BLE_CONNECTION_ERROR,
    BLE_CONNECTION_TIMEOUT,
    BLE_DISCONNECTION_ERROR,
    BLE_INTERNAL_SOFTWARE_ERROR,
    BLE_DEVICE_ERROR,
    BLE_NOTIF_ERROR,
    BLE_READ_READY,
    BLE_READ_WAIT,
    BLE_READ_TIMEOUT,
    BLE_READ_ERROR,
    BLE_DISCOVER_ERROR
} ble_group_status_t;

typedef enum ble_sensor_state {
    SENSOR_FOUND = 0,
    SENSOR_NOT_FOUND
} ble_sensor_state_t;

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            STRUCT
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

typedef struct ble_group_s ble_group_t;
typedef struct ble_sensor_s ble_sensor_t;
typedef struct BLE_FunctionCbS ble_function_ptr_t;
typedef struct userdata_s userdata_t;

struct ble_sensor_s {
  const char *uid;
  const char *info;
  char *macAddr;
  const char *apiverb;
  ble_group_t * group;
  BLEMode_t mode;
  json_object *sample;
  json_object *version;
  gatt_connection_t * gatt_connection;
  BLEState_t state;
  int connectingCount;
  int noDataCount;
  bool found;
  pthread_t threadID;
  afb_api_t api;
  afb_event_t event;
  afb_event_t eventShock;
  void * ctx;
};

struct ble_group_s {
  const char *uid;
  const char *info;
  const char *prefix;
  const char *adminapi;
  int nbrSensors;
  ble_sensor_t *sensors;
  TimerHandleT *timer;
  ble_function_ptr_t *functions;
};

struct userdata_s {
  ble_group_t *groups;
  int nbrGroups;
  TimerHandleT *timerState;
};

struct BLE_FunctionCbS {
  const char *uid;
  const char *info;
  int (*init_ptr_func) (ble_sensor_t *sensor, json_object *argJ);
  int (*config_ptr_func)(ble_sensor_t *sensor, json_object *argJ);
  int (*version_ptr_func)(ble_sensor_t *sensor/*, json_object **outputJ*/);
  int (*data_streaming_ptr_func)(ble_sensor_t *sensor);
};


typedef struct BLE_gattlib_ble_sensor_s {
    const char *macAddr;    
    char *name;
    ble_sensor_state_t state;
} BLE_gattlib_ble_sensor_t;

/**
 * ble_sensor_thread_s - Hold usefull information for thread
 */
typedef struct ble_sensor_thread_s{
    pthread_t threadID;
    afb_api_t api;
    void *ctx;
} ble_sensor_thread_t;

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            FUNCTION TYPE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

typedef void (*register_function_ptr_t) (const char *uid, ble_function_ptr_t *functionCB);

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                    BLE GATTLIB INTERFACE FUNCTIONS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

/**
 * ble_gattlib_disconnect_handler - Handler for device disconnected
 *
 * @param user_data userData
 * @return status: 0 ion success
 */
void ble_gattlib_disconnect_handler(void * user_data);

/**
 * ble_gattlib_connect_async_cb - Callback call when we try to connect in async
 *
 * @param connection gatt connection of the sensor
 * @param user_data userData
 */
void ble_gattlib_connect_async_cb(gatt_connection_t* connection, void* user_data);

/**
 * ble_gattlib_connect - Search the sensor and connect to it
 *
 * @param ble_sensor_t *sensor: information of the sensor
 * @param json_object * argsJ: macAddress of the sensor in json string
 * @param ble_group_status_t *status: Status of the process
 * @param BLEMode_t asyncMode: BLE_ASYNC_MODE or BLE_NON_ASYNC_MODE
 * @return 0 in success, negative int in error
 */
ble_group_status_t ble_gattlib_connect (ble_sensor_t *sensor, json_object * argsJ, BLEMode_t asyncMode);

/**
 * ble_gattlib_disconnect - disconnect to the sensor
 *
 * @param gatt_connection_t * connection: gattlib connection struct
 * @return ble_group_status_t
 */
ble_group_status_t ble_gattlib_disconnect (gatt_connection_t * connection);

/**
 * ble_gattlib_discover - Get all char and services of the sensor
 *
 * @param ble_sensor_t *sensor
 * @param json_object ** outputJ: Response of the request
 * @return ble_group_status_t
 */
ble_group_status_t ble_gattlib_discover(ble_sensor_t *sensor, json_object ** outputJ);

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                    BLE SENSOR GLUE FUNCTIONS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

/**
 * ble_glue_sensors_info - Create Json in respons to the info verb
 *      of the api of the ble
 * 
 * @param ble ptr of ble_group_t
 * @param verbose level of the verbosity
 * @param responsJ Response in Json to use
 * @return 0 if success, negative int if not
 */
int ble_glue_sensors_info(ble_group_t * ble, int verbose, json_object * responseJ);

/**
 * ble_glue_group_request - Callback to the api/verb of a BLE section
 * 
 * @param request the api request
 * @param ble ptr to ble group data
 * @param queryJ The query in Json
 */
void ble_glue_group_request(afb_req_t request, ble_group_t * group, json_object *queryJ);

/**
 * ble_glue_sensor_request - Callback to the api/verb of a Sensor
 * 
 * @param request the api request
 * @param sensor ptr of ble_sensor_t
 * @param queryJ The query in Json
 */
void ble_glue_sensor_request(afb_req_t request, ble_sensor_t * sensor, json_object *queryJ);

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                    BLE SENSOR CALLBACK FUNCTIONS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

/**
 * bleCallbackRegister - Register pointer function func of a plugin for a sensor
 *  Data stored into global variable
 * 
 * @param uid Uid of the plugin
 * @param funcs List of ble_function_ptr_t
 */
void ble_function_ptr_register (char *uid, ble_function_ptr_t *funcs);

/**
 * ble_function_ptr_find - Find the correct ble_function_ptr_t for BLE_group with an uid
 * 
 * @param api afb_api_t
 * @param uid:The uid of the BLE
 * @return the ble_function_ptr_t found
 */
ble_function_ptr_t * ble_function_ptr_find (afb_api_t api, const char *uri);

#endif /*_BLE_SENSOR_BINDING_INCLUDE_*/