/*
 * Copyright 2020 - 2021 IoT.bzh
 *
 * author: Valentin Lefebvre <valentin.lefebvre@iot.bzh>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            DEFINE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

#define _GNU_SOURCE

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            INCLUDE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

#include "ble-sensor-binding.h"

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            STRUCT
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

typedef struct ble_registry_s {
   char *uid;
   struct ble_registry_s *next;
   ble_function_ptr_t *functionsPtr;
} ble_registry_t;

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            GLOBAL VARIABLE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

// registry holds a linked list of core+pugins encoders
static ble_registry_t *registryHead = NULL;

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            PRIVATE FUNCTION
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

/**
 * PluginarseUID - Parse the uid from the data raw
 *  Example of uri: "plugin://uid#format"
 * 
 * @param uri the URI in raw
 * @param plugin uid of the plugin parsed
 * @param format Format of the plugin parsed
 */
static void PluginParseUID (const char* uri, char **plugin, char **format) {
    char *chaine;
    int idx;
    static int prefixlen = sizeof(PLUGIN_ACTION_PREFIX)-1;

    if (strncasecmp(uri, PLUGIN_ACTION_PREFIX, prefixlen)) goto NoPluginExit;

    // break URI in substring ignoring leading tcp:
    chaine= strdup(uri);   
    for (idx=prefixlen; idx < strlen(chaine); idx ++) {
        if (chaine[idx] == 0) break;
        if (chaine[idx] == '#') chaine[idx] = 0;
    }

    // extract IP addr and port
    *plugin = &chaine[prefixlen];
    *format= &chaine[idx];
    return;

NoPluginExit:
    return;
}

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            STATIC FUNCTION
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

/**
 * bleFunctionFind - Find specific functions from a ble_function_ptr_t
 * 
 * @param functions the ble_function_ptr_t list where to search the specific function
 * @param uid The function uid to find
 * @return the ble_function_ptr_t found
 */
static ble_function_ptr_t *bleFunctionFind (ble_function_ptr_t *functions, const char *uid) {
    int idx;
    assert (uid);
    assert(functions);

    for (idx=0; functions[idx].uid; idx++) {
        if (!strcasecmp (functions[idx].uid, uid)) break;
    }

    return (&functions[idx]);
}

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            PUBLIC FUNCTION
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

void ble_function_ptr_register (char *uid, ble_function_ptr_t *bleFuncs) {
    ble_registry_t * registryIdx, * registryEntry;

    registryEntry = (ble_registry_t *) calloc(1, sizeof(ble_registry_t));
    registryEntry->uid = uid;
    registryEntry->functionsPtr = bleFuncs;

    // if not 1st encoder insert at the end of the chain
    if (!registryHead) {
        registryHead = registryEntry;
    } else {
        for (registryIdx = registryHead; registryIdx->next; registryIdx=registryIdx->next);
        registryIdx->next = registryEntry;
    }
}

ble_function_ptr_t * ble_function_ptr_find (afb_api_t api, const char *uid) {
    char *pluginuid = NULL, *formatuid = NULL;
    ble_registry_t *registryIdx;
    ble_function_ptr_t *functionsPtr;


    PluginParseUID (uid, &pluginuid, &formatuid);

    if (!pluginuid) goto OnErrorExit;

    for (registryIdx= registryHead; registryIdx->next; registryIdx=registryIdx->next) {
        if (registryIdx->uid && !strcasecmp (registryIdx->uid, uid)) break;
    }
    if (!registryIdx) {
        AFB_API_ERROR(api, "ble_function_ptr_find: Fail to find plugin='%s' format bleFuncs ", formatuid);
        goto OnErrorExit;
    }

    functionsPtr = bleFunctionFind(registryIdx->functionsPtr, formatuid);
    if (!functionsPtr || !functionsPtr->uid) {
        AFB_API_ERROR(api, "ble_function_ptr_find: Fail to find plugin='%s' format='%s' bleFuncs", pluginuid, formatuid);
        goto OnErrorExit;
    } 

    return (functionsPtr);

OnErrorExit:
    return NULL;
}