/*
 * Copyright 2020 - 2021 IoT.bzh
 *
 * author: Valentin Lefebvre <valentin.lefebvre@iot.bzh>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            DEFINE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

#define _GNU_SOURCE

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            INCLUDE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

#include "ble-sensor-binding.h"

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            PRIVATE FUNCTIONS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

/**
 * SensorEventCreate - Call when subscribe is called, to create the event
 * 
 * @param sensor ptr of Sensor
 * @return 0 if success, negative int if not 
 */
static int SensorEventCreate(ble_sensor_t * sensor) {
    ble_group_t * ble = sensor->group;
    char eventName[32], eventShockName[32];
    int ret = 0 ;
    ret = snprintf(eventName, sizeof(eventName), "%s/%s", ble->uid, sensor->uid);
    ret |= snprintf(eventShockName, sizeof(eventShockName), "%s/%s/shock", ble->uid, sensor->uid);
    if (ret < 0) {
        AFB_API_ERROR(sensor->api, "[SensorEventCreate] error on snprintf");
        return 1;
    }

    // If no event create one an attach it to the sensor
    if (! afb_event_is_valid(sensor->event)) {
        sensor->event = afb_api_make_event(sensor->api, eventName);
        if (!sensor->event) goto MakeEventError;
    }
    if (! afb_event_is_valid(sensor->eventShock)) {
        sensor->eventShock = afb_api_make_event(sensor->api, eventShockName);
        if (!sensor->eventShock) goto MakeEventError;
    }
    return 0;

MakeEventError:
    AFB_API_ERROR(sensor->api, "SensorEventCreate: fail to create event ble=%s sensor=%s",
            ble->uid, sensor->uid);
    goto OnErrorExit;

OnErrorExit:
    return -1;
}

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            PUBLIC FUNCTIONS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

int ble_glue_sensors_info(ble_group_t * ble, int verbose, json_object * responseJ) {
    json_object *elemJ = NULL;
    json_object *actionsJ = NULL;
    json_object *usageJ = NULL;
    ble_sensor_t *sensor;
    int ret = 0;
    char *verb_uid;

    // Loop on every Sensor
    for (int index = 0; index < ble->nbrSensors; index++) {
        sensor = &ble->sensors[index];
        ret = asprintf(&verb_uid, "%s/%s", ble->uid, sensor->uid);
        if (ret < 0) {
            AFB_ERROR("[ble_glue_sensors_info] - asprintf error for ble %s/%s", ble->uid, ble->sensors[index].uid);
            goto OnExit;
        }else {
            switch (verbose) {
                default:
                case 1 :
                    ret = wrap_json_pack(&elemJ, "{ss ss ss}", 
                            "uid", sensor->uid,
                            "family", ble->uid,
                            "info", sensor->info);
                    break;
                case 2 :
                    ret = wrap_json_pack(&elemJ, "{ss ss ss ss ss}", 
                            "uid", sensor->uid,
                            "family", ble->uid,
                            "info", sensor->info,
                            "macAddr", sensor->macAddr,
                            "verb", verb_uid);
                    break;
                case 3 :
                    if (sensor->sample) json_object_get(sensor->sample);

                    ret = wrap_json_pack(&actionsJ, "[s s s s s s s s]", "connect", "disconnect", "discover", "config", "subscribe", "unsubscribe", "version", "state");
                    if (ret < 0) {
                        AFB_ERROR("[ble_glue_sensors_info] Fail to wrap actionJ for %s/%s", ble->uid, sensor->uid);
                        goto OnExit;
                    }
                    ret = wrap_json_pack (&usageJ, "{so*, ss}", "action", actionsJ, "data", "test");
                    if (ret < 0) {
                        AFB_ERROR("[ble_glue_sensors_info] Fail to wrap usageJ for %s/%s", ble->uid, sensor->uid);
                        goto OnExit;
                    }

                    ret = wrap_json_pack(&elemJ, "{ss ss ss ss ss so* so*}", 
                            "uid", sensor->uid,
                            "family", ble->uid,
                            "info", sensor->info,
                            "macAddr", sensor->macAddr,
                            "verb", verb_uid,
                            "usage", usageJ,
                            "sample", sensor->sample);
                    break;
            }

            if (ret < 0 ) {
                AFB_ERROR("[ble_glue_sensors_info] Fail to wrap json ble info ble=%s sensor=%s", ble->uid, sensor->uid);
                goto OnExit;
            } else {
                json_object_array_add (responseJ, elemJ);
            }
        }
    }

OnExit:
    return ret;
}

void ble_glue_group_request(afb_req_t request, ble_group_t * group, json_object *queryJ) {
    int err;
    ble_group_status_t ret = 0;
    const char *action;
    int verbose=0;
    json_object *responseJ=NULL;

    err= wrap_json_unpack(queryJ, "{ss s?s !}", "action", &action, "verbose", &verbose);
    if (err < 0) {
        afb_req_fail_f (request, "Failed", "invalid query BLE=%s query=%s", group->uid, json_object_get_string(queryJ));
        return;
    }

    if (!strcasecmp(action, "connect")) {

        // Loop on every Sensor
        for (int index = 0; index < group->nbrSensors; index++) {
            // Create the connection
            ret = ble_gattlib_connect(&group->sensors[index], queryJ, BLE_SENSOR_NON_ASYNC_MODE);
            if (ret != BLE_SUCCESS) {
                afb_req_fail_f(request, "connect error", "ble_glue_group_request: Failed to connect to the sensor with return=%i", ret);
                return;
            }

            AFB_API_NOTICE(request->api, "Sensor with addr=%s is connected",group->sensors[index].macAddr);
            group->sensors[index].state = BLE_SENSOR_CONNECTED_STATE;
        }

    }else if (!strcasecmp(action, "disconnect")) {
        // Loop on every Sensor
        for (int index = 0; index < group->nbrSensors; index++) {
            ret = 0;
            if (group->sensors[index].state != BLE_SENSOR_CONNECTED_STATE) break;

            if (! (&(group->sensors[index]))->gatt_connection) {
                afb_req_fail_f(request, "disconnect error", "ble_glue_group_request: Failed to disconnect - not implemented");
                return;
            }

            // Disconnect the sensor
            ret = ble_gattlib_disconnect((&(group->sensors[index]))->gatt_connection);
            if (ret != BLE_SUCCESS) {
                afb_req_fail_f(request, "disconnect error", "ble_glue_group_request: Failed to disconnect from the sensor with return=%i", ret);
                return ;
            }

            group->sensors[index].state = BLE_SENSOR_DISCONNECTED_STATE;
        }

    } else if (!strcasecmp(action, "info")){

        responseJ = json_object_new_array();
        ret = ble_glue_sensors_info(group, verbose, responseJ);
        if (ret < 0) {
            afb_req_fail_f(request, "Info error", "ble_glue_group_request Fail to create the response for groups %s", group->uid);
            json_object_put(responseJ);
            return;
        }

    } else {
        afb_req_fail_f(request, "Syntax error", "ble_glue_group_request: Action=%s is unknown", action);
        return;
    }

    afb_req_success(request, responseJ, NULL);
    return;
}

void ble_glue_sensor_request(afb_req_t request, ble_sensor_t * sensor, json_object *queryJ) {
    assert(sensor);
    assert(queryJ);
    
    int err = 0;
    ble_group_status_t ret = BLE_SUCCESS;
    ble_group_t * ble =  sensor->group;
    const char * action;
    json_object * dataJ, * responseJ = NULL;


    err= wrap_json_unpack(queryJ, "{ss s?o !}", "action", &action, "data", &dataJ);
    if (err < 0) {
        afb_req_fail_f(request, "query-error", "ble_glue_sensor_request: invalid 'json' ble=%s sensor=%s query=%s"
            , ble->uid, sensor->uid, json_object_get_string(queryJ));
        return;
    }

    AFB_API_INFO(afb_req_get_api(request), "calling verb: %s for sensor %s/%s", action, ble->uid, sensor->uid);

    /*%%%%% Connect action %%%%%*/
    if (!strcasecmp(action, "CONNECT")) {

        if (sensor->state != BLE_SENSOR_DISCONNECTED_STATE) {
            afb_req_fail_f(request, "Already Connected", "ble_glue_sensor_request: The sensor %s is already connected", sensor->uid);
            return;
        }
        // Create the connection
        // sensor->gatt_connection = connect_sensor_ble(sensor, dataJ, &ret, BLE_SENSOR_NON_ASYNC_MODE);
        ret = ble_gattlib_connect(sensor, queryJ, BLE_SENSOR_NON_ASYNC_MODE);
        if (ret < 0) {
            AFB_API_WARNING(request->api, "Sensor with addr=%s cannot be connect", sensor->macAddr);
            afb_req_fail_f(request, "connect error", "ble_glue_sensor_request: fail to connect to the sensor=%s/%s with ret = %i, please retry!",
                ble->uid, sensor->uid, ret);
            return;   
        }
        AFB_API_NOTICE(request->api, "Sensor with addr=%s is connected", sensor->macAddr);

    /*%%%%% Disconnect action %%%%%*/
    } else if (!strcasecmp(action, "DISCONNECT")) {

        if (sensor->state == BLE_SENSOR_DISCONNECTED_STATE) goto NotConnected;
        if (!sensor->gatt_connection) goto NotConnected;

        // Disconnect the sensor
        ret = ble_gattlib_disconnect(sensor->gatt_connection);
        if (ret != BLE_SUCCESS) {
            afb_req_fail_f(request, "disconnect error", "ble_glue_sensor_request: fail to disconnect from ble=%s sensor=%s",
                ble->uid, sensor->uid);
            return;
        }
        sensor->state = BLE_SENSOR_DISCONNECTED_STATE;

    /*%%%%% Subscribe action %%%%%*/
    } else if (!strcasecmp(action, "SUBSCRIBE")) {

        // Check the state
        if (sensor->state == BLE_SENSOR_DISCONNECTED_STATE) goto NotConnected;
        // Create an event if no one exists already
        err = SensorEventCreate(sensor);
        if (err < 0) {
            AFB_API_ERROR(request->api, "[Subscribe] - Cannot create the event to subscribe for sensor %s/%s", ble->uid, sensor->uid);
            afb_req_fail_f(request, "subscribe error","Internal error - Cannot create event");
            return;
        }
        // Subscribe to this events
        err = afb_req_subscribe(request, sensor->event);
        if (err < 0) {
            afb_req_fail_f(request, "subscribe error","ble_glue_sensor_request: fail to subscribe on sensor %s/%s event - error=%i",
                ble->uid, sensor->uid, err);
            return;
        }
        err = afb_req_subscribe(request, sensor->eventShock);
        if (err < 0) {
            afb_req_fail_f(request, "subscribe error","ble_glue_sensor_request: fail to subscribe on sensor %s/%s shocks - error=%i",
                ble->uid, sensor->uid, err);
            return;
        }

    /*%%%%% Unsubscribe action %%%%%*/
    } else if (!strcasecmp(action, "UNSUBSCRIBE")) {

        if (! afb_event_is_valid(sensor->event)) {
            afb_req_fail_f(request, "unsubscribe error","ble_glue_sensor_request: fail to unsubscribe event of sensor %s/%s is not valid",
                ble->uid, sensor->uid);
            return;
        }
        err = afb_req_unsubscribe(request, sensor->event);
        if (err < 0) {
            afb_req_fail_f(request, "unsubscribe error","ble_glue_sensor_request: fail to unsubscribe event of sensor %s/%s",
                ble->uid, sensor->uid);
            return;
        }
        err = afb_req_unsubscribe(request, sensor->eventShock);
        if (err < 0) {
            afb_req_fail_f(request, "unsubscribe error","ble_glue_sensor_request: fail to unsubscribe shocks event of sensor %s/%s",
                ble->uid, sensor->uid);
            return;
        }

    /*%%%%% Discover action %%%%%*/
    } else if (!strcasecmp(action, "DISCOVER")) {

        if (sensor->state == BLE_SENSOR_DISCONNECTED_STATE) goto NotConnected;
        ret = ble_gattlib_discover(sensor, &responseJ);
        if (ret != BLE_SUCCESS) {
            afb_req_fail_f(request, "discover error","ble_glue_sensor_request: fail to discover services/characteristics for sensor=%s/%s",
                ble->uid, sensor->uid);
            return;
        }
    
    /*%%%%% Config action %%%%%*/
    } else if (!strcasecmp(action, "CONFIG")) {

        if (sensor->state == BLE_SENSOR_DISCONNECTED_STATE) goto NotConnected;
        // Call Callback's sensor function of Configuration
        if(!ble->functions->config_ptr_func) {
            afb_req_fail_f(request, "config warning","ble_glue_sensor_request: Set config function unavailable for %s/%s",
                ble->uid, sensor->uid);
            return;
        }
        err = (ble->functions->config_ptr_func) (sensor, dataJ);
        if (err < 0) {
            afb_req_fail_f(request, "config error","ble_glue_sensor_request: fail to configure ble=%s sensor=%s",
                ble->uid, sensor->uid);
            return;
        }
        
    /*%%%%% Version action %%%%%*/
    } else if (!strcasecmp(action, "VERSION")) {

        if (sensor->state == BLE_SENSOR_DISCONNECTED_STATE) goto NotConnected;
        if (!sensor->version) {
            afb_req_fail_f(request, "version warning","ble_glue_sensor_request: Version unavailable for %s/%s",
                ble->uid, sensor->uid);
            return;
        }
        afb_req_success(request, sensor->version, NULL);
        json_object_get(sensor->version);
        return;
    
    /*%%%%% State action %%%%%*/
    } else if (!strcasecmp(action, "STATE")) {
        switch (sensor->state)
        {
        case BLE_SENSOR_CONNECTED_STATE:
            wrap_json_pack(&responseJ, "{ss*}", "state", "connected");
            break;
        
        case BLE_SENSOR_CONNECTING_STATE:
            wrap_json_pack(&responseJ, "{ss*}", "state", "connecting...");
            break;
        
        case BLE_SENSOR_DISCONNECTED_STATE:
            wrap_json_pack(&responseJ, "{ss*}", "state", "disconnected");
            break;
        
        case BLE_SENSOR_CONFIG_STATE:
            wrap_json_pack(&responseJ, "{ss*}", "state", "in configuration ...");
            break;
        
        case BLE_SENSOR_STREAMING_STATE:
            wrap_json_pack(&responseJ, "{ss*}", "state", "data streaming");
            break;

        default:
            wrap_json_pack(&responseJ, "{ss*}", "state", "unknown");
            break;
        }

        switch (sensor->mode)
        {
        case BLE_SENSOR_MANUAL_MODE:
            json_object_object_add(responseJ, "mode", json_object_new_string("manual"));
            break;
        
        case BLE_SENSOR_AUTO_MODE:
            json_object_object_add(responseJ, "mode", json_object_new_string("auto"));
            break;

        default:
            json_object_object_add(responseJ, "mode", json_object_new_string("unknown"));
            break;
        }
    } else {
        goto OnUnknownActionError;
    }

    afb_req_success(request, responseJ, NULL);
    return;

NotConnected:
    afb_req_fail_f(request, "Not Connected", "The sensor %s is not connected",
            sensor->uid);
    return;

OnUnknownActionError:
    afb_req_fail_f(request, "syntax-error","ble_glue_sensor_request: action=%s UNKNOWN ble=%s sensor=%s error=%i",
            action, ble->uid, sensor->uid, err);
    return;

// OnErrorDisconnect:
//     ret = ble_gattlib_disconnect(sensor->gatt_connection);
//     if (ret != BLE_SUCCESS) {
//         AFB_API_ERROR(sensor->api, "Cannot disconnect the sensor -> please restart the service");
//         return;
//     }
//     sensor->state = BLE_SENSOR_DISCONNECTED_STATE;
//     return;
}