/*
 * Copyright (C) 2020 - 2021 "IoT.bzh"
 *
 * Author: Valentin Lefebvre <valentin.lefebvre@iot.bzh>
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _GNU_SOURCE

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            INCLUDE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

#include "ble-sensor-binding.h"
#include <ctl-plugin.h>

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            DEFINE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

/* 
 * Constant value for all the UUIDs
 * ---------------------------------
 *  Note : we need to define "short" uuid strings because gattlib only takes
 *            the first part of the uuid
 *
 *  For further informations about UUIDs see (page 4-5):
 *  https://www.bosch-connectivity.com/media/downloads/ciss/ciss_ble_communication_protocol.pdf
 */
#define INERTIAL_STR_UUID       "00007502-0000-1000-8000-00805f9b34fb"
#define INERTIAL_SHORT_STR_UUID "0x7502"
#define ENVIRONMENTAL_STR_UUID  "00007504-0000-1000-8000-00805f9b34fb"
#define ENVIRONMENTAL_SHORT_STR_UUID "0x7504"

// #define DETECTION_CONFIG_STR_UUID  "00007701-0000-1000-8000-00805f9b34fb"
// #define DETECTION_CONFIG_SHORT_STR_UUID "0x7701"

// #define DETECTION_EVENT_STR_UUID  "00007702-0000-1000-8000-00805f9b34fb"
// #define DETECTION_EVENT_SHORT_STR_UUID "0x7702"

#define CONFIG_STR_UUID         "0000750a-0000-1000-8000-00805f9b34fb"

#define MANUFACTURER_STR_UUID   "00002a29-0000-1000-8000-00805f9b34fb"
#define MODEL_NBR_STR_UUID      "00002a24-0000-1000-8000-00805f9b34fb"
#define SERIAL_NBR_STR_UUID     "00002a25-0000-1000-8000-00805f9b34fb"
#define HARD_REV_STR_UUID       "00002a27-0000-1000-8000-00805f9b34fb"
#define FIRM_REV_STR_UUID       "00002a26-0000-1000-8000-00805f9b34fb"
#define SOFT_REV_STR_UUID       "00002a28-0000-1000-8000-00805f9b34fb"
#define SYS_ID_STR_UUID         "00002a23-0000-1000-8000-00805f9b34fb"
#define DEVICE_NAME_STR_UUID    "00002a00-0000-1000-8000-00805f9b34fb"
#define DEVICE_INFO_STR_UUID    "0000180a-0000-1000-8000-00805f9b34fb"

#define SHOCK_ACCX_DETECTION 4 // in G
#define SHOCK_ACCY_DETECTION 4 // in G
#define SHOCK_ACCZ_DETECTION 3 // in G

/* 
 * Constant For Redis dataBase
 * ----------------------------
 */
#define DTB_API "redis"
#define DTB_INSERT "ts_jinsert"
#define DTB_CLASS "ciss"
#define DTB_CLASS_SHOCK "ciss_shock"

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            STRUCT
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

/**
 * ciss_bosh_config_s - Represante last Data sending by the CISS Bosh
 *  For each tab [Minimum, Averrage, Maximum]
 * 
 *  temp: Temperature Minimum, Maximum, Average
 *  hum: Humidity Minimum, Maximum, Average
 *  press: Pressure Minimum, Maximum, Average
 *  accX: Accelerometer (x, y, z)
 *  gyrX: Gyrometer (x, y, z)
 *  magX: Magnometer (x, y, z)
 */
typedef struct ciss_bosh_data_s {
    float temp[3];
    float hum[3];
    float press[3];
    int16_t accX[4]; // Fourth value is for the last data catch: useful for algo
    int16_t accY[4]; // Fourth value is for the last data catch: useful for algo
    int16_t accZ[4]; // Fourth value is for the last data catch: useful for algo
    uint16_t gyrX[3];
    uint16_t gyrY[3];
    uint16_t gyrZ[3];
    uint16_t magX[3];
    uint16_t magY[3];
    int16_t magZ[3];
    ssize_t countEnv;
    ssize_t countIner;
}ciss_bosh_data_t;

/**
 * ciss_bosh_config_s -  Represante the sensor Config
 * 
 *  samplingRate: 0x01->10ms,  0x01->10ms, 0x02->100ms, 0x03->1000ms, 0x04->10000ms, 0x05->30000ms, 0x06->60000ms, 0x07->600000ms
 *  dataSelection: 0 (disable), 1 (enable)
 *  blePowerLevel: level0 -> -18, level1 -> -15, level2 -> -12, level3 -> -9, level4 -> -6, level5 -> -3, level6 -> 0 (default), level7 -> +3 in dBm
 */
typedef struct ciss_bosh_config_s {
    uint8_t samplingRate;
    uint8_t dataSelection;
    uint8_t blePowerLevel;
}ciss_bosh_config_t;

/**
 * Struct: ciss_bosh_uuid_s - Represant Some usefull uuid to communicate with CISS BOSH sensor
 */
typedef struct ciss_bosh_uuid_s {
    uuid_t config_uuid;
    uuid_t environmental_uuid;
    uuid_t inertial_uuid;
    uuid_t detection_event_uuid;
    uuid_t detection_config_uuid;
    uuid_t manufacturer_uuid;
    uuid_t model_nbr_uuid;
    uuid_t serial_nbr_uuid;
    uuid_t hard_rev_uuid;
    uuid_t firm_rev_uuid;
    uuid_t soft_rev_uuid;
    uuid_t sys_id_uuid;
    uuid_t device_name_uuid;
    uuid_t device_info_uuid;
}ciss_bosh_uuid_t;


typedef struct ciss_bosh_info_s {
    char *manufacture;
    char *model_nbr;
    char *serial_nbr;
    char *hard_rev;
    char *firm_rev;
    char *soft_rev;
    char *sys_id;
    char *device_name;
    char *device_info;
}ciss_bosh_info_t;

/**
 * ciss_bosh_ctx_s - Specific context of the CISS BOSH sensor
 *  Content all uuids and configurations
 */
typedef struct ciss_bosh_ctx_s {
    ciss_bosh_info_t info;
    ciss_bosh_data_t data;
    ciss_bosh_uuid_t uuids;
    ciss_bosh_config_t config;
    int samplingEnvRedis;
    int samplingInertRedis;
    TimerHandleT *timerEnvRedis;
    TimerHandleT *timerInertRedis;
    bool dataInertReceived;
    bool dataEnvReceived;
}ciss_bosh_ctx_t;

/**
 * kind_event - all kinds of potential event
 */
typedef enum kind_event {
    EVENT_ENVIRONMENT=0,
    EVENT_INERTIAL,
    EVENT_SHOCK
}kind_event_t;

/**
 * ciss_bosh_timerEvt_s - Hold usefull information for Timer handler
 */
typedef struct ciss_bosh_timerEvt_s {
    ble_sensor_t *sensor;
    kind_event_t kindEvt;
}ciss_bosh_timerEvt_t;

/**
 * sensor_thread_s - Hold usefull information for thread
 */
typedef struct sensor_thread_s{
    pthread_t threadID;
    ble_sensor_t *sensor;
} sensor_thread_t;

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            GLOBAL VARIABLE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

pthread_mutex_t lock;
int shock_lock = 0;

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            STATIC FUNCTIONS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

CTLP_CAPI_REGISTER("ciss-bosh")

/**
 * decode_environment_data - Decode environmental data and store into context
 *  For further informations see (page 7-8):
 *  https://www.bosch-connectivity.com/media/downloads/ciss/ciss_ble_communication_protocol.pdf
 *
 * @param sensor sensor data / context
 * @param const uint8_t *buffer: the buffer of data sent by the sensor
 */
static void decode_environment_data(ble_sensor_t *sensor, const uint8_t *buffer)
{
    float temperature = (float) (((buffer[0] << 8) | buffer[1]) / 10.0);
    float humidity = (float) ((buffer[2] << 8 | buffer[3]) / 100.0);
    float pressure = (float) (((buffer[4] << 24) | (buffer[5] << 16) | (buffer[6] << 8) | buffer[7]) / 100.0);

    pthread_mutex_lock(&lock);
    ciss_bosh_ctx_t * ctx = (ciss_bosh_ctx_t *) sensor->ctx;
    ciss_bosh_data_t * data = &((ciss_bosh_ctx_t *) sensor->ctx)->data;
    ciss_bosh_data_t old_sensor_data = ctx->data;

    /* Store into Sensor context */
    // Temperature
    data->temp[0] = (old_sensor_data.temp[0] < temperature && old_sensor_data.temp[0] != 0)? old_sensor_data.temp[0] : temperature;
    data->temp[1] += temperature;
    data->temp[2] = (old_sensor_data.temp[2] > temperature)? old_sensor_data.temp[2] : temperature;
    // Humidity
    data->hum[0] = (old_sensor_data.hum[0] < humidity && old_sensor_data.hum[0] != 0)? old_sensor_data.hum[0] : humidity;
    data->hum[1] += humidity;
    data->hum[2] = (old_sensor_data.hum[2] > humidity)? old_sensor_data.hum[2] : humidity;
    // Pressure
    data->press[0] = (old_sensor_data.press[0] < pressure && old_sensor_data.press[0] != 0)? old_sensor_data.press[0] : pressure;
    data->press[1] += pressure;
    data->press[2] = (old_sensor_data.press[2] > pressure)? old_sensor_data.press[2] : pressure;

    data->countEnv ++;
    pthread_mutex_unlock(&lock);
    return;
}

/**
 * cbk - Empty callback for afb_api_call not to SEGFAULT
 */
void cbk(void *closure, struct json_object *obj, const char *err, const char *info, afb_api_t api){
    kind_event_t kind_event = *(kind_event_t *) closure;
    switch (kind_event)
    {
    case EVENT_ENVIRONMENT:
        AFB_API_INFO(api, "Environmental data has been send to Redis");
        break;
    
    case EVENT_INERTIAL:
        AFB_API_INFO(api, "Inertial data has been send to Redis");
        break;
    
    case EVENT_SHOCK:
        AFB_API_INFO(api, "Shock event has been send to Redis");
        break;
    }

    return;
}

/**
 * unlockShockEventCB - Unlock the shock mutex
 */
static void unlockShockEventCB(int signal, void * ctx) {
    // Signal should be null
    if(signal) return;
    shock_lock = 0;
    return;
}

/**
 * decode_inertial_data - Decode inertial data and store into context
 *  For further informations see (page 7):
 *  https://www.bosch-connectivity.com/media/downloads/ciss/ciss_ble_communication_protocol.pdf
 * 
 * @param sensor sensor data / context
 * @param const uint8_t *buffer: the buffer of data sent by the sensor
 */
static void decode_inertial_data(ble_sensor_t *sensor, const uint8_t *buffer/*, json_object **outputJ*/)
{
    int err;
    // int count;
    json_object * shockEventJ = NULL, * reqJ = NULL;
    // json_object * accJ, * gyrJ, * magJ;
    int16_t shockX = 0, shockY = 0, shockZ = 0;
    kind_event_t kind_event_shock = EVENT_SHOCK;

    int16_t ax = (int16_t) ((buffer[0] << 8) | buffer[1]);
    int16_t ay = (int16_t) ((buffer[2] << 8) | buffer[3]);
    int16_t az = (int16_t) ((buffer[4] << 8) | buffer[5]);
    uint16_t gx = (uint16_t) (((buffer[6] << 8) | buffer[7]) >> 4);
    uint16_t gy = (uint16_t) (((buffer[7] & 0x0f) << 8) | buffer[8]);
    uint16_t gz = (uint16_t) ((buffer[9] << 4) | (buffer[10] >> 4));
    uint16_t mx = (uint16_t) (((((buffer[10] & 0x0f) << 8) | buffer[11]) << 2) | (buffer[12] >> 6));
    uint16_t my = (uint16_t) (((buffer[12] & 0x3f) << 8) | buffer[13]);
    int16_t mz = (uint16_t) ((buffer[14] << 8) | buffer[15]);

    pthread_mutex_lock(&lock);
    ciss_bosh_ctx_t * ctx = (ciss_bosh_ctx_t *) sensor->ctx;
    ciss_bosh_data_t * data = &((ciss_bosh_ctx_t *) sensor->ctx)->data;
    ciss_bosh_data_t old_sensor_data = ctx->data;

    /* Detection of Shocks */
    if (shock_lock != 1) {
        shockX = (int16_t) (data->accX[3] - ax);
        shockY = (int16_t) (data->accY[3] - ay);
        shockZ = (int16_t) (data->accZ[3] - az);
        if (shockX >= 1000*SHOCK_ACCX_DETECTION || shockX <= -1000*SHOCK_ACCX_DETECTION ||
                shockY >= 1000*SHOCK_ACCY_DETECTION || shockY <= -1000*SHOCK_ACCY_DETECTION ||
                shockZ >= 1000*SHOCK_ACCZ_DETECTION || shockZ <= -1000*SHOCK_ACCZ_DETECTION ) {
            AFB_API_INFO(sensor->api, "Shock apears !!!");
            err = wrap_json_pack(&shockEventJ, "{si si si}",
                    "Acceleration X", shockX,
                    "Acceleration Y", shockY,
                    "Acceleration Z", shockZ);
            if(err) AFB_API_WARNING(sensor->api, "decode_inertial_data: Failed to parse JSON shock detection");

            AFB_API_DEBUG(sensor->api, "Shock : %s", json_object_get_string(shockEventJ));

            // Push event if someone subscribe to it
            if (afb_event_is_valid(sensor->eventShock)) {
                afb_event_push(sensor->eventShock, json_object_get(shockEventJ));
            }

            // Send shock to redis
            wrap_json_pack(&reqJ, "{s:s s:o}",
                "class", DTB_CLASS_SHOCK,
                "data", shockEventJ);
            afb_api_call(sensor->api, DTB_API, DTB_INSERT, reqJ, cbk, &kind_event_shock);

            // Lock the shock process to avoid an overload of data send
            err = afb_api_queue_job(afbBindingRoot, unlockShockEventCB, sensor, NULL, -5000);
            if (err <= 0) {
                AFB_API_WARNING(sensor->api, "Failed to schedule the jobs that will unlock shock event");
            } else {
                shock_lock = 1;
            }
        }
    }
    /* Store into Sensor Context */
    // Accelerometer
    data->accX[0] = (old_sensor_data.accX[0] < ax && old_sensor_data.accX[0] != 0)? old_sensor_data.accX[0] : ax;
    data->accX[1] = (int16_t) (data->accX[1] + ax);
    data->accX[2] = (old_sensor_data.accX[2] > ax)? old_sensor_data.accX[2] : ax;
    data->accX[3] = ax;
    
    data->accY[0] = (old_sensor_data.accY[0] < ay && old_sensor_data.accY[0] != 0)? old_sensor_data.accY[0] : ay;
    data->accY[1] = (int16_t) (data->accY[1] + ay);
    data->accY[2] = (old_sensor_data.accY[2] > ay)? old_sensor_data.accY[2] : ay;
    data->accY[3] = ay;

    data->accZ[0] = (old_sensor_data.accZ[0] < az && old_sensor_data.accZ[0] != 0)? old_sensor_data.accZ[0] : az;
    data->accZ[1] = (int16_t) (data->accZ[1] + az);
    data->accZ[2] = (old_sensor_data.accZ[2] > az)? old_sensor_data.accZ[2] : az;
    data->accZ[3] = az;

    // Gyroscope
    data->gyrX[0] = (old_sensor_data.gyrX[0] < gx && old_sensor_data.gyrX[0] != 0)? old_sensor_data.gyrX[0] : gx;
    data->gyrX[1] = (int16_t) (data->gyrX[1] + gx);
    data->gyrX[2] = (old_sensor_data.gyrX[2] > gx)? old_sensor_data.gyrX[2] : gx;

    data->gyrY[0] = (old_sensor_data.gyrY[0] < gy && old_sensor_data.gyrY[0] != 0)? old_sensor_data.gyrY[0] : gy;
    data->gyrY[1] = (int16_t) (data->gyrY[1] + gy);
    data->gyrY[2] = (old_sensor_data.gyrY[2] > gy)? old_sensor_data.gyrY[2] : gy;

    data->gyrZ[0] = (old_sensor_data.gyrZ[0] < gz && old_sensor_data.gyrZ[0] != 0)? old_sensor_data.gyrZ[0] : gz;
    data->gyrZ[1] = (int16_t) (data->gyrZ[1] + gz);
    data->gyrZ[2] = (old_sensor_data.gyrZ[2] > gz)? old_sensor_data.gyrZ[2] : gz;

    // Magnometer
    data->magX[0] = (old_sensor_data.magX[0] < mx && old_sensor_data.magX[0] != 0)? old_sensor_data.magX[0] : mx;
    data->magX[1] = (int16_t) (data->magX[1] + mx);
    data->magX[2] = (old_sensor_data.magX[2] > mx)? old_sensor_data.magX[2] : mx;

    data->magY[0] = (old_sensor_data.magY[0] < my && old_sensor_data.magY[0] != 0)? old_sensor_data.magY[0] : my;
    data->magY[1] = (int16_t) (data->magY[1] + my);
    data->magY[2] = (old_sensor_data.magY[2] > my)? old_sensor_data.magY[2] : my;

    data->magZ[0] = (old_sensor_data.magZ[0] < mz && old_sensor_data.magZ[0] != 0)? old_sensor_data.magZ[0] : mz;
    data->magZ[1] = (int16_t) (data->magZ[1] + mz);
    data->magZ[2] = (old_sensor_data.magZ[2] > mz)? old_sensor_data.magZ[2] : mz;

    data->countIner ++;
    pthread_mutex_unlock(&lock);

    return;
}

/**
 * CissTimerCallback - CAllback on timer to check news data stored from the CISS Bosh sensor
 * 
 * @param timer timer handle
 */
static int CissTimerCallback (TimerHandleT *timer) {
    assert(timer);

    ciss_bosh_timerEvt_t * timerCtx = (ciss_bosh_timerEvt_t *) timer->context;
    ble_sensor_t * sensor = timerCtx->sensor;

    if (sensor->state != BLE_SENSOR_STREAMING_STATE) {
        // Sensor is not configurate yet, so do nothing
        if (sensor->state == BLE_SENSOR_DISCONNECTED_STATE) {
            AFB_API_DEBUG(sensor->api, "TIMER CALLBACK -> disconnected mode => STOP TIMER");
            // Sensor has been disconnected 
            return 0;
        }
        return 1;
    }

    ciss_bosh_data_t * data = &((ciss_bosh_ctx_t *) sensor->ctx)->data;
    json_object * dataJ = NULL, * tempJ = NULL, * humJ = NULL, * pressJ = NULL, * accJ = NULL, * gyrJ = NULL, * magJ = NULL, * reqJ = NULL;
    float averTemp, averHum, averPress;
    int16_t averAccX, averAccY, averAccZ;
    uint16_t averGyrX, averGyrY, averGyrZ;
    uint16_t averMagX, averMagY;
    int16_t averMagZ;
    int count;

    pthread_mutex_lock(&lock);

    switch (timerCtx->kindEvt)
    {
    case EVENT_ENVIRONMENT:
        if (data->countEnv == 0) goto NoDataCaught;
        sensor->noDataCount = 0;
        averTemp =  (float) (data->temp[1]/(float)data->countEnv);
        averHum =  (float) (data->hum[1]/(float)data->countEnv);
        averPress =  (float) (data->press[1]/(float)data->countEnv);
        wrap_json_pack(&tempJ, "{sf sf sf}", "Min", data->temp[0], "Max", data->temp[2], "Average", averTemp);
        wrap_json_pack(&humJ, "{sf sf sf}", "Min", data->hum[0], "Max", data->hum[2], "Average", averHum);
        wrap_json_pack(&pressJ, "{sf sf sf}", "Min", data->press[0], "Max", data->press[2], "Average", averPress);
        wrap_json_pack(&dataJ, "{so so so}", "Temperature", tempJ, "Humidity", humJ, "Pressure", pressJ);

        AFB_API_DEBUG(sensor->api, "Environmental (count: %li): %s", data->countEnv, json_object_get_string(dataJ));

        /* Clear env data */ 
        memset(data->temp, 0, sizeof(data->temp));
        memset(data->hum, 0, sizeof(data->temp));
        memset(data->press, 0, sizeof(data->temp));
        data->countEnv = 0;
        break;
    
    case EVENT_INERTIAL:
        if (data->countIner == 0) goto NoDataCaught;
        sensor->noDataCount = 0;
        averAccX = (int16_t) (data->accX[1]/data->countIner);
        averAccY = (int16_t) (data->accY[1]/data->countIner);
        averAccZ = (int16_t) (data->accZ[1]/data->countIner);
        averGyrX = (uint16_t) (data->gyrX[1]/data->countIner);
        averGyrY = (uint16_t) (data->gyrY[1]/data->countIner);
        averGyrZ = (uint16_t) (data->gyrZ[1]/data->countIner);
        averMagX = (uint16_t) (data->magX[1]/data->countIner);
        averMagY = (uint16_t) (data->magY[1]/data->countIner);
        averMagZ = (int16_t) (data->magZ[1]/data->countIner);
        wrap_json_pack(&accJ, "{si si si si si si si si si}", 
                    "x-min", data->accX[0], "x-max", data->accX[2], "x-average", averAccX,
                    "y-min", data->accY[0], "y-max", data->accY[2], "y-average", averAccY,
                    "z-min", data->accZ[0], "z-max", data->accZ[2], "z-average", averAccZ);
        wrap_json_pack(&gyrJ, "{si si si si si si si si si}", 
                    "x-min", data->gyrX[0], "x-max", data->gyrX[2], "x-average", averGyrX,
                    "y-min", data->gyrY[0], "y-max", data->gyrY[2], "y-average", averGyrY,
                    "z-min", data->gyrZ[0], "z-max", data->gyrZ[2], "z-average", averGyrZ);
        wrap_json_pack(&magJ, "{si si si si si si si si si}", 
                    "x-min", data->magX[0], "x-max", data->magX[2], "x-average", averMagX,
                    "y-min", data->magY[0], "y-max", data->magY[2], "y-average", averMagY,
                    "z-min", data->magZ[0], "z-max", data->magZ[2], "z-average", averMagZ);
        wrap_json_pack(&dataJ, "{so so so}", "Accelerometer", accJ, "Gyroscope", gyrJ, "Magnometer", magJ);

        AFB_API_DEBUG(sensor->api, "Inertial (count: %li): %s", data->countIner, json_object_get_string(dataJ));

        /* Clear Inertial data */ 
        memset(data->accX, 0, sizeof(data->accX));
        memset(data->accY, 0, sizeof(data->accY));
        memset(data->accZ, 0, sizeof(data->accZ));
        memset(data->gyrX, 0, sizeof(data->gyrX));
        memset(data->gyrY, 0, sizeof(data->gyrY));
        memset(data->gyrZ, 0, sizeof(data->gyrZ));
        memset(data->magX, 0, sizeof(data->magX));
        memset(data->magY, 0, sizeof(data->magY));
        memset(data->magZ, 0, sizeof(data->magZ));
        data->countIner = 0;

        break;

    default:
        AFB_API_WARNING(timer->api, "CissTimerCallback: Wrong type of Event");
        break;
    }
    pthread_mutex_unlock(&lock);

    // Push event if someone subscribe to it
    if (afb_event_is_valid(sensor->event)) {
        count = afb_event_push(sensor->event, json_object_get(dataJ));
        // dataJ still needed to send to redis so use this function below to not unref the data
        if (count == 0) {
            AFB_API_DEBUG(timer->api, "No more client listen event");
        }
    }

    // Send data to Redis database
    wrap_json_pack(&reqJ, "{s:s s:o}",
            "class", DTB_CLASS,
            "data", dataJ);
    afb_api_call(timer->api, DTB_API, DTB_INSERT, reqJ, cbk, &timerCtx->kindEvt);

    return 1; // Returning 0 would stop the timer

NoDataCaught:
    pthread_mutex_unlock(&lock);
    
    if (sensor->noDataCount > 2) {
        AFB_API_WARNING(timer->api, "No Data Catches from the device for a while ->disconnect the sensor");
        sensor->state = BLE_SENSOR_ERROR_STATE;
        sensor->noDataCount = 0;
        ((ciss_bosh_ctx_t *) sensor->ctx)->dataEnvReceived = false;
        ((ciss_bosh_ctx_t *) sensor->ctx)->dataInertReceived = false;
        TimerEvtStop(((ciss_bosh_ctx_t *) sensor->ctx)->timerEnvRedis);
        TimerEvtStop(((ciss_bosh_ctx_t *) sensor->ctx)->timerInertRedis);
        return 0;
        //TimerEvtStop(timer);

        // ciss_bosh_ctx_t *sensor_ctx = (ciss_bosh_ctx_t *) sensor_ctx;

    } else {
        AFB_API_WARNING(timer->api, "No Data Catches from the device -> Sending nothing to Redis");
        sensor->noDataCount ++;
    }
    return 1;
}

/**
 * notification_handler - Callback function call when a notification is received, it
 *  checks the corresponding uuid and call the proper decoding function
 *
 * @param afb_api_t api: the service api
 * @param const uuid_t *uuid: the uuid that push the notification
 * @param const uint8_t *data: data array push by the sensor
 * @param size_t data_length: size of the data push by the sensor
 * @param void *user_data: custom data sended (if any), these data are
 *      parametrized when you registered this callback
 *      function
 */
void notification_handler(const uuid_t *uuid, const uint8_t *data, size_t data_length, void *user_data)
{
    (void)data_length;
    assert(user_data);
    char uuid_str[MAX_LEN_UUID_STR + 1] = {0};
    ble_sensor_t * sensor = (ble_sensor_t *) user_data;
    ciss_bosh_ctx_t * sensor_ctx = (ciss_bosh_ctx_t *) sensor->ctx;

    // Convert raw uuid to string uuid
    if (gattlib_uuid_to_string(uuid, uuid_str, sizeof(uuid_str)) != GATTLIB_SUCCESS) {
        AFB_WARNING("notification_handler: Cannot decode UUID incoming ...");
        return;
    };
    
    if (strcmp(uuid_str, INERTIAL_SHORT_STR_UUID) == 0) {
        sensor_ctx->dataInertReceived = true;
        decode_inertial_data(sensor, data);
    } else if (strcmp(uuid_str, ENVIRONMENTAL_SHORT_STR_UUID) == 0) {
        sensor_ctx->dataEnvReceived = true;
        decode_environment_data(sensor, data);
    }
    return;
}

/**
 * initialize_uuids - Transform string uuid into raw uuids
 *
 * @param afb_api_t api: the service api
 * @param ciss_bosh_uuid_t *ciss_bosh_uuid: raw uuid pointer of confif data
 * @return ble_group_status_t
 */
static ble_group_status_t initialize_uuids(afb_api_t api, ciss_bosh_uuid_t *ciss_bosh_uuid)
{
    if (gattlib_string_to_uuid(INERTIAL_STR_UUID, UUID_LENGTH + 1, &ciss_bosh_uuid->inertial_uuid) != GATTLIB_SUCCESS ||
            gattlib_string_to_uuid(ENVIRONMENTAL_STR_UUID, UUID_LENGTH + 1, &ciss_bosh_uuid->environmental_uuid) != GATTLIB_SUCCESS ||
            gattlib_string_to_uuid(CONFIG_STR_UUID, UUID_LENGTH + 1, &ciss_bosh_uuid->config_uuid) != GATTLIB_SUCCESS ||
            gattlib_string_to_uuid(MANUFACTURER_STR_UUID, UUID_LENGTH +1, &ciss_bosh_uuid->manufacturer_uuid) != GATTLIB_SUCCESS ||
            gattlib_string_to_uuid(MODEL_NBR_STR_UUID, UUID_LENGTH +1, &ciss_bosh_uuid->model_nbr_uuid) != GATTLIB_SUCCESS ||
            gattlib_string_to_uuid(SERIAL_NBR_STR_UUID, UUID_LENGTH +1, &ciss_bosh_uuid->serial_nbr_uuid) != GATTLIB_SUCCESS ||
            gattlib_string_to_uuid(HARD_REV_STR_UUID, UUID_LENGTH +1, &ciss_bosh_uuid->hard_rev_uuid) != GATTLIB_SUCCESS ||
            gattlib_string_to_uuid(FIRM_REV_STR_UUID, UUID_LENGTH +1, &ciss_bosh_uuid->firm_rev_uuid) != GATTLIB_SUCCESS ||
            gattlib_string_to_uuid(SOFT_REV_STR_UUID, UUID_LENGTH +1, &ciss_bosh_uuid->soft_rev_uuid) != GATTLIB_SUCCESS ||
            gattlib_string_to_uuid(SYS_ID_STR_UUID, UUID_LENGTH +1, &ciss_bosh_uuid->sys_id_uuid) != GATTLIB_SUCCESS ||
            gattlib_string_to_uuid(DEVICE_NAME_STR_UUID, UUID_LENGTH +1, &ciss_bosh_uuid->device_name_uuid) != GATTLIB_SUCCESS ||
            gattlib_string_to_uuid(DEVICE_INFO_STR_UUID, UUID_LENGTH +1, &ciss_bosh_uuid->device_info_uuid) != GATTLIB_SUCCESS/* ||
            gattlib_string_to_uuid(DETECTION_CONFIG_STR_UUID, UUID_LENGTH +1, &ciss_bosh_uuid->detection_config_uuid) != GATTLIB_SUCCESS ||
            gattlib_string_to_uuid(DETECTION_EVENT_STR_UUID, UUID_LENGTH +1, &ciss_bosh_uuid->detection_event_uuid) != GATTLIB_SUCCESS*/) {
        AFB_API_ERROR(api, "initialize_uuids: Fail to convert string uuid to string, exiting ...");
        return -BLE_CONFIG_ERROR;
    }
    
    return BLE_SUCCESS;
}

/** 
 * monitorStreamingThread - Take care about the config to CISS Bash sensor
 * to start the streaming data 
 */
static void *monitorStreamingThread (void *ctx) {
    sensor_thread_t *threadCtx = (sensor_thread_t *) ctx;
    ble_sensor_t * sensor = threadCtx->sensor;
    ciss_bosh_ctx_t *sensor_ctx = (ciss_bosh_ctx_t *) sensor->ctx;
    int ret = 0;

    AFB_API_INFO(sensor->api, "[thread Streaming %s/%s] setup Ciss", sensor->group->uid, sensor->uid);

    // Here we want to enable all sensor and being notified every second
    uint8_t data_config[20];
    data_config[0] = 0x19; /* bit0: 1 to enable data sending by BLE // bit3: Enable Frequency changing // bit4: Enable BLE Transmit Power level change */
    data_config[1] = 0x0f; /* Enable all return data */
    data_config[2] = sensor_ctx->config.samplingRate;
    data_config[3] = sensor_ctx->config.blePowerLevel;
    for (int index = 4; index <20; index ++) {
        data_config[index] = 0x00;
    }
    ret = gattlib_write_char_by_uuid(sensor->gatt_connection, &sensor_ctx->uuids.config_uuid, data_config, sizeof(data_config));
    if ( ret != GATTLIB_SUCCESS) {
        AFB_API_ERROR(sensor->api, "[thread Streaming] Cannot write chars to set configuration, error : %i", ret);
        goto OnExit;
    }

    gattlib_register_notification(sensor->gatt_connection, notification_handler, (void *)sensor);

    // sleep(5);

    if (!sensor_ctx->dataInertReceived) {

        AFB_API_INFO(sensor->api, "[thread Streaming %s/%s] start inert notif", sensor->group->uid, sensor->uid);
        ret = gattlib_notification_start(sensor->gatt_connection, &sensor_ctx->uuids.inertial_uuid);
        AFB_API_DEBUG(sensor->api, "[%s %s/%s] return of notification start", __func__, sensor->group->uid, sensor->uid);
        if (ret != GATTLIB_SUCCESS) {
            AFB_API_ERROR(sensor->api, "[thread Streaming] Error in starting inertial notif: %i", ret);
            goto OnExit;
        }
        uint8_t data_inertial[20];
        data_inertial[0] = 0x01;
        data_inertial[1] = 0x00;
        for (int index = 2; index <20; index ++) {
            data_inertial[index] = 0x00;
        }
        ret = gattlib_write_char_by_uuid(sensor->gatt_connection, &sensor_ctx->uuids.inertial_uuid, data_inertial, sizeof(data_inertial));
        if ( ret != GATTLIB_SUCCESS) {
            AFB_API_ERROR(sensor->api, "[thread Streaming] Cannot write chars to set configuration, error : %i", ret);
            goto OnExit;
        }
    } else {
        AFB_API_DEBUG(sensor->api, "no need to start inert notif");
    }
    // //usleep(50000);

    if (!sensor_ctx->dataEnvReceived) {

        AFB_API_INFO(sensor->api, "[thread Streaming %s/%s] start env notif", sensor->group->uid, sensor->uid);
        ret = gattlib_notification_start(sensor->gatt_connection, &sensor_ctx->uuids.environmental_uuid);
        if (ret != GATTLIB_SUCCESS) {
            AFB_API_ERROR(sensor->api, "[notifStartCB] Error in starting environmental notif: %i", ret);
            goto OnExit;
        }
        uint8_t data_env[20];
        data_env[0] = 0x01;
        data_env[1] = 0x00;
        for (int index = 2; index <20; index ++) {
            data_env[index] = 0x00;
        }
        ret = gattlib_write_char_by_uuid(sensor->gatt_connection, &sensor_ctx->uuids.environmental_uuid, data_env, sizeof(data_env));
        if ( ret != GATTLIB_SUCCESS) {
            AFB_API_ERROR(sensor->api, "[thread Streaming] Cannot write chars to set configuration, error : %i", ret);
            goto OnExit;
        }
    }  else {
        AFB_API_DEBUG(sensor->api, "no need to start env notif");
    }

    AFB_API_INFO(sensor->api, "[thread Streaming %s/%s] streaming started !", sensor->group->uid, sensor->uid);
    sensor->state = BLE_SENSOR_STREAMING_STATE;
    free(threadCtx);
    pthread_exit(NULL);

OnExit:
    sensor->state = BLE_SENSOR_ERROR_STATE;
    free(threadCtx);
    pthread_exit(NULL);
}

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            CALLBACKS API FUNCTIONS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

/**
 * ciss_bosh_init - Initiate CISS Bosh connection
 * 
 * @param sensor sensor info / context
 * @param argsJ dynamics information to apply
 * @return 0 if success or negative if error
 */
static int ciss_bosh_init(ble_sensor_t *sensor, json_object * argsJ) {
    int ret = 0;
    ciss_bosh_ctx_t * sensor_ctx;
    ciss_bosh_uuid_t sensor_uuids;
    ciss_bosh_config_t sensor_config;
    ciss_bosh_data_t sensor_data ={0};
    sensor_ctx = (ciss_bosh_ctx_t *) calloc(2, sizeof(ciss_bosh_ctx_t));

    // Initialize all necessary uuids
    ret = initialize_uuids(sensor->api, &sensor_uuids);
    if (ret != BLE_SUCCESS) {
        AFB_API_ERROR(sensor->api, "[ciss_bosh_init] Fail to convert string uuid to string for sensor= %s/%s _ error: %i", 
            sensor->group->uid, sensor->uid, ret);
        goto OnExit;
    }
 
    // Catch Sampling for Redis DB
    ret = wrap_json_unpack(argsJ, "{si si si si si}",
            "samplingRedisEnv", &sensor_ctx->samplingEnvRedis,
            "samplingRedisIner", &sensor_ctx->samplingInertRedis,
            "samplingRate", &sensor_config.samplingRate,
            "dataSelection", &sensor_config.dataSelection,
            "blePowerLevel", &sensor_config.blePowerLevel);
    if (ret < 0) {
        AFB_API_ERROR(sensor->api, "[ciss_bosh_init] Fail to unwrap sampleting info for sensor= %s/%s", 
            sensor->group->uid, sensor->uid);
        goto OnExit;
    }

    // Adding to the context
    sensor_ctx->data = sensor_data;
    sensor_ctx->uuids = sensor_uuids;
    sensor_ctx->config = sensor_config;
    sensor_ctx->dataInertReceived = false;
    sensor_ctx->dataEnvReceived = false;
    sensor->ctx = (void *)sensor_ctx;

    // Initiate the mutex
    if(pthread_mutex_init(&lock, NULL) != 0) {
        AFB_API_ERROR(sensor->api, "[ciss_bosh_init] Fail to initiate the mutex for sensor= %s/%s", 
            sensor->group->uid, sensor->uid);
        return -1;
    }

OnExit:
    return ret;
}

/**
 * ciss_bosh_dataStreaming - Start the data streaming process
 * 
 * @param sensor sensor info / context
 * @return 0 if success or negative if error
 */
static int ciss_bosh_dataStreaming(ble_sensor_t *sensor) {
    ciss_bosh_ctx_t * sensor_ctx = (ciss_bosh_ctx_t *) sensor->ctx;
    char * timeruid;
    ciss_bosh_timerEvt_t * cissEnvEvtHandle, * cissInertEvtHandle;
    int ret = 0;

    // Do not process if it is not in connected state
    if( sensor->state != BLE_SENSOR_CONNECTED_STATE) {
        AFB_API_ERROR(sensor->api, "[ciss_bosh_dataStreaming] The sensor (%s/%s) is not in the good state", sensor->group->uid, sensor->uid);
        return -BLE_CONFIG_ERROR;
    }
    sensor->state = BLE_SENSOR_CONFIG_STATE;
    sensor_thread_t * threadCtx = calloc(1, sizeof(sensor_thread_t));
    threadCtx->sensor = sensor;
    ret = pthread_create(&threadCtx->threadID, NULL, monitorStreamingThread, threadCtx);
    if (ret < 0) {
        AFB_API_ERROR(sensor->api, "[ciss_bosh_dataStreaming] Error to create streaming thread");
        return -1;
    }

    // Start Timer to send Env data to Redis database
    sensor_ctx->timerEnvRedis = (TimerHandleT*)calloc (1, sizeof(TimerHandleT));
    sensor_ctx->timerEnvRedis->delay = (uint) 1000 * sensor_ctx->samplingEnvRedis;
    sensor_ctx->timerEnvRedis->count = -1; // run forever
    asprintf (&timeruid, "%s/%s/environment", sensor->group->uid, sensor->uid);
    sensor_ctx->timerEnvRedis->uid = timeruid;

    cissEnvEvtHandle = (ciss_bosh_timerEvt_t *) calloc(1, sizeof(ciss_bosh_timerEvt_t));
    cissEnvEvtHandle->sensor = sensor;
    cissEnvEvtHandle->kindEvt = EVENT_ENVIRONMENT;
    TimerEvtStart (sensor->api, sensor_ctx->timerEnvRedis, CissTimerCallback, (void *)cissEnvEvtHandle);

    // Start Timer to send Env data to Redis database
    sensor_ctx->timerInertRedis = (TimerHandleT*)calloc (1, sizeof(TimerHandleT));
    sensor_ctx->timerInertRedis->delay = (uint) 1000 * sensor_ctx->samplingInertRedis;
    sensor_ctx->timerInertRedis->count = -1; // run forever
    asprintf (&timeruid, "%s/%s/Inertial", sensor->group->uid, sensor->uid);
    sensor_ctx->timerEnvRedis->uid = timeruid;

    cissInertEvtHandle = (ciss_bosh_timerEvt_t *) calloc(1, sizeof(ciss_bosh_timerEvt_t));
    cissInertEvtHandle->sensor = sensor;
    cissInertEvtHandle->kindEvt = EVENT_INERTIAL;
    TimerEvtStart (sensor->api, sensor_ctx->timerInertRedis, CissTimerCallback, cissInertEvtHandle);

    return BLE_SUCCESS;
}

/**
 * ciss_bosh_config - configurate the CISS Bosh sensor thanks to args
 * 
 * @param sensor sensor info / context
 * @param argJ usefull info to configure the sensor
 * @return 0 if success or negative if error
 */
static int ciss_bosh_config(ble_sensor_t *sensor, json_object *argJ) {
    assert(argJ);
    int ret = 0, sampling = 0, dataSelect = 0, blePower = 0;
    ciss_bosh_ctx_t * ctx = (ciss_bosh_ctx_t *) sensor->ctx;
    ciss_bosh_config_t sensor_config;

    ret = wrap_json_unpack(argJ, "{s?i s?i s?i}",
            "samplingRate", &sampling,
            "dataSelection", &dataSelect,
            "blePowerLevel", &blePower);
    if (ret < 0) {
        AFB_API_ERROR(sensor->api, "ciss_bosh_config: Fail to unpack the JSON for the sensor=%s/%s", sensor->group->uid, sensor->uid);
        goto OnExit;
    }

    sensor_config.samplingRate = (uint8_t) sampling; 
    sensor_config.dataSelection = (uint8_t) dataSelect; 
    sensor_config.blePowerLevel = (uint8_t) blePower; 
    ctx->config = sensor_config;

OnExit:
    return ret;
}

typedef struct {
    ciss_bosh_ctx_t * sensor_ctx;
    uuid_t *characteristic_uuid;
}info_cb_data_t;

void *_read_info_cb (const void *buffer, size_t buffer_len, void * user_data) {
    if (user_data == NULL)
        return NULL;

    info_cb_data_t * data = (info_cb_data_t *) user_data;
    if (data->characteristic_uuid == NULL) {
        AFB_WARNING("[%s] characteristic uuid is NULL !!", __func__);
        return NULL;
    }

    char uuid_str[MAX_LEN_UUID_STR + 1];
    gattlib_uuid_to_string(data->characteristic_uuid, uuid_str, MAX_LEN_UUID_STR + 1);

    if (!gattlib_uuid_cmp(data->characteristic_uuid, &data->sensor_ctx->uuids.manufacturer_uuid)) {
        asprintf(&data->sensor_ctx->info.manufacture, "%s", (char *) buffer);
    }
    if (!gattlib_uuid_cmp(data->characteristic_uuid, &data->sensor_ctx->uuids.model_nbr_uuid)) {
        asprintf(&data->sensor_ctx->info.model_nbr, "%s", (char *) buffer);
    }
    if (!gattlib_uuid_cmp(data->characteristic_uuid, &data->sensor_ctx->uuids.serial_nbr_uuid)) {
        asprintf(&data->sensor_ctx->info.serial_nbr, "%s", (char *) buffer);
    }
    if (!gattlib_uuid_cmp(data->characteristic_uuid, &data->sensor_ctx->uuids.hard_rev_uuid)) {
        asprintf(&data->sensor_ctx->info.hard_rev, "%s", (char *) buffer);
    }
    if (!gattlib_uuid_cmp(data->characteristic_uuid, &data->sensor_ctx->uuids.firm_rev_uuid)) {
        asprintf(&data->sensor_ctx->info.firm_rev, "%s", (char *) buffer);
    }
    if (!gattlib_uuid_cmp(data->characteristic_uuid, &data->sensor_ctx->uuids.soft_rev_uuid)) {
        data->sensor_ctx->info.soft_rev = (char *) calloc(buffer_len + 1, sizeof(char));
        for (int index = 0; index < buffer_len; index ++) {
            sprintf(data->sensor_ctx->info.soft_rev+index, "%d", ((uint8_t *)buffer)[index]);
        }
    }
    if (!gattlib_uuid_cmp(data->characteristic_uuid, &data->sensor_ctx->uuids.sys_id_uuid)) {
        asprintf(&data->sensor_ctx->info.sys_id, "%s", (char *) buffer);
    }
    if (!gattlib_uuid_cmp(data->characteristic_uuid, &data->sensor_ctx->uuids.device_name_uuid)) {
        asprintf(&data->sensor_ctx->info.device_name, "%s", (char *) buffer);
    }
    if (!gattlib_uuid_cmp(data->characteristic_uuid, &data->sensor_ctx->uuids.device_info_uuid)) {
        asprintf(&data->sensor_ctx->info.device_info, "%s", (char *) buffer);
    }
    
    return NULL;
}

/**
 * ciss_bosh_version - ask version to CISS Bosh sensor
 * 
 * @param sensor sensor info / context
 * @param outputJ response in JSON for info action
 * @return 0 if success or negative if error
 */
static int ciss_bosh_version(ble_sensor_t *sensor/*, json_object **outputJ*/) {
    int ret =0;

    // Check state of pointers
    if (!sensor->gatt_connection) {
        AFB_API_ERROR(sensor->api, "[ciss_bosh_version] The Gatt connection of the sensor=%s/%s is NULL", sensor->group->uid, sensor->uid);
        ret = -BLE_NOT_CONNECTED;
        goto OnExit;
    }
    if (sensor->state == BLE_SENSOR_DISCONNECTED_STATE || sensor->state == BLE_SENSOR_STREAMING_STATE) {
        AFB_API_ERROR(sensor->api, "[ciss_bosh_version] sensor =%s/%s is in the wrong state", sensor->group->uid, sensor->uid);
        ret = -BLE_WRONG_STATE;
        goto OnExit;
    }
    assert(sensor->ctx);
    ciss_bosh_ctx_t * sensor_ctx = (ciss_bosh_ctx_t *) sensor->ctx;

    uuid_t * uuid_ciss[9] = {0};
    uuid_ciss[0] = &sensor_ctx->uuids.manufacturer_uuid;
    uuid_ciss[1] = &sensor_ctx->uuids.model_nbr_uuid;
    uuid_ciss[2] = &sensor_ctx->uuids.serial_nbr_uuid;
    uuid_ciss[3] = &sensor_ctx->uuids.hard_rev_uuid;
    uuid_ciss[4] = &sensor_ctx->uuids.firm_rev_uuid;
    uuid_ciss[5] = &sensor_ctx->uuids.soft_rev_uuid;
    uuid_ciss[6] = &sensor_ctx->uuids.sys_id_uuid;
    uuid_ciss[7] = &sensor_ctx->uuids.device_name_uuid;
    uuid_ciss[8] = &sensor_ctx->uuids.device_info_uuid;

    for (int idx =0; idx < 9; idx++) {
        char uuid_str[256];
        gattlib_uuid_to_string(uuid_ciss[idx], uuid_str, sizeof(uuid_str));
        info_cb_data_t cb_data = {
            .characteristic_uuid = uuid_ciss[idx],
            .sensor_ctx = sensor_ctx,
        };
        ret = gattlib_read_char_by_uuid_async(sensor->gatt_connection, uuid_ciss[idx], _read_info_cb, (void *)&cb_data);
        // To not be blocked on while loop, we need to use async and wait a delay to be sure to get the information ...
        usleep(500000); // 500ms
    }

    ret = wrap_json_pack(&(sensor->version), "{s?s* s?s* s?s* s?s* s?s* s?s* s?s* s?s* s?s*}",
            "Manufacturer", sensor_ctx->info.manufacture,
            "Model Number", sensor_ctx->info.model_nbr,
            "Serial Number", sensor_ctx->info.serial_nbr,
            "Hardware Revision", sensor_ctx->info.hard_rev,
            "Firmware Revision", sensor_ctx->info.firm_rev,
            "Software Revision", sensor_ctx->info.soft_rev,
            "System ID", sensor_ctx->info.sys_id,
            "Device Name", sensor_ctx->info.device_name,
            "Device Info", sensor_ctx->info.device_info);

    if (ret < 0) {
        AFB_API_ERROR(sensor->api, "ciss_bosh_version: Fail to pack version in JSON for the sensor=%s/%s", 
            sensor->group->uid, sensor->uid);
        ret = -1;
    }
OnExit:
    return ret;
}

// functions callbacks
static ble_function_ptr_t cissSensor[] = {
    {.uid="saft", .info="Plugins for CISS Bosh sensor", 
        .init_ptr_func=ciss_bosh_init,
        .data_streaming_ptr_func=ciss_bosh_dataStreaming,
        .config_ptr_func=ciss_bosh_config,
        /*.version_ptr_func=ciss_bosh_version*/},
    {.uid=NULL} // must be NULL terminated
};

CTLP_ONLOAD(plugin, functionPtr) {
    register_function_ptr_t func = (register_function_ptr_t) functionPtr;
    assert(func);
    (*func) (plugin->uid, cissSensor);
    return 0;
}